/*
*       Copyright (c) 2011-2022 by esd electronics gmbh
*
*  This software is copyrighted by and is the sole property of
*  esd gmbh.  All rights, title, ownership, or other interests
*  in the software remain the property of esd gmbh. This
*  software may only be used in accordance with the corresponding
*  license agreement.  Any unauthorized use, duplication, transmission,
*  distribution, or disclosure of this software is expressly forbidden.
*
*  This Copyright notice may not be removed or modified without prior
*  written consent of esd gmbh.
*
*  esd gmbh, reserves the right to modify this software without notice.
*
*  esd electronics gmbh                   Tel. +49-511-37298-0
*  Vahrenwalder Strasse 207               Fax. +49-511-37298-68
*  30165 Hannover                         http://www.esd.eu
*  Germany                                sales@esd.eu
*
*************************************************************************
*
* Description:
* =============================================================================
* This module contains complex slave example code. Slave has an object
* dictionary (OD) and some of its items are mappable into process data.
* This app shows how the stack API to access these variables are used, as
* well as some of the possible callbacks and flags.
*
* When an output variable is changed, it's printed.
* Input variables are changed every second (when in "SafeOp" or "Op")
*
* Remember: Used slave code, slave ESI and Slave EEPROM content must all match!
*
*
* RxPDOs (Outputs):
*     0x1600 [Assigned by default]
*         Contains Obj. 0x2000: "Output1", 32 bit
*     0x1601
*         Contains Obj. 0x2001: "Output2", 32 bit
*
* TxPDOs (Inputs):
*     0x1a00 [Assigned by default]
*         Contains Obj. 0x2010: "Input1", 32 bit
*     0x1a01
*         Contains Obj. 0x2011: "Input2", 32 bit
*         Contains Obj. 0x2012: "Input3", 8 bit
*
*
* Object dictionary, automatically added items:
***********************************************
*   Objects 0x1000, 0x1008, 0x1009, 0x100a, 0x1018
*     by essODAddGenericObjects()
*   Object 0x1c00
*     by flag ESS_OD_FLAGS_HANDLE_SM_TYPES
*   Objects 0x1c12, 0x1c13
*     by essODUpdatePDOAssignment()
*   Objects 0x1600, 0x1601, 0x1a00, 0x1a01
*     by essODUpdatePDOConfiguration()
*
* Object dictionary, manually added items:
******************************************
*   Objects 0x2000, 0x2001
*       Output variables
*
*   Objects 0x2010, 2011, 2012
*       Input variables
*
*   Object 0x2100: "Test Object", a record with different entries
*       1: "TickCount", Read-only, "data" parameter for this entry is "NULL"
*           i.e. data is acquired by callback, see cbCoEReadWrite()
*       2: "Test-String", RW entry, data also acquired by callback
*          When uploading/reading each byte will be the segment number
*          When downloading/writing this entry the data is only dumped, see
*          cbCoEReadWrite()
*       3: "Test-OctetString", same as entry 2, just of type octet-string
*       4: "TestEntry1", RW entry, 8 bit unsigned integer
*          Has "Min", "Max" and "Default" value given (not actually checked
*          here, just to be displayed by configuration tool, etc.)
*       5: "TestEntry2", RW entry, 8 bit signed integer
*       6: "TestEntry3", RW entry, 16 bit unsigned integer
*       7: "TestEntry4", RW entry, 16 bit signed integer
*       8: "TestEntry5", RW entry, 32 bit unsigned integer
*       9: "TestEntry6", RW entry, 32 bit signed integer
*      10: "Test-Emergency", W entry, 16 bit unsigned integer
*          When writing, it creates an emergency.
*
*/




/* Includes: */
/*************/
#include "ess.h" /* includes essSystem.h which usually includes all standard stuff like <stdio.h> etc. */


/* Application/Slave configuration macros: */
/*******************************************/

#ifndef ESS_COMPLEX_PD_WATCHDOG                         /* Set to 1 to let the app switch back to SafeOp when no outputs received from master for some time */
# define ESS_COMPLEX_PD_WATCHDOG                        1
#endif

#ifndef ESS_COMPLEX_VERBOSE                             /* Set to 0 to disable some outputs */
# define ESS_COMPLEX_VERBOSE                            1
#endif

#ifndef ESS_COMPLEX_SHOW_PDO_CONFIG                     /* Only for info, see ShowPDOAssignmentAndConfig() */
# define ESS_COMPLEX_SHOW_PDO_CONFIG                    1
#endif

    /*
     * If the slave shall support the "Explicit Device ID" mechanism "Direct ID"
     * defined in ETG.1020 you can configure the address where the ID is copied
     * to, below. The value shall be in the memory ranges 0x0F18..0x0F1F or
     * 0x1000..0x1003.
     */
#ifndef ESS_COMPLEX_DIRECT_ID_ADO                       
# define ESS_COMPLEX_DIRECT_ID_ADO                      0x0
#endif

#ifndef ESS_COMPLEX_ID_RELOAD_OBJECT                    /* Set to 1 to support the ID Reload Object according to ETG.1020 */
# define ESS_COMPLEX_ID_RELOAD_OBJECT                   1
#endif

#ifndef ESS_COMPLEX_MAIN                                /* Might be used when this file is just included into a project where main() already exists */
# define ESS_COMPLEX_MAIN                               main /* or OS like VxWorks where main() is not allowed for a loadable kernel module */
#endif

/* _MEASURE defines below use slave's DC timestamps - keep in mind that this clock is the one whose speed is/can be adapted! */

#ifndef ESS_COMPLEX_JITTER_MEASURE                      /* Set to 1 to let the app estimate jitter of output events */
# define ESS_COMPLEX_JITTER_MEASURE                     0
#endif

#ifndef ESS_COMPLEX_IRQDELAY_MEASURE                    /* Set to 1 to let the app estimate interrupt delay by SYNC0 event (Needs SYNC0 enabled at high interval (e.g. 1 second) and mapped to Al event request) */
# define ESS_COMPLEX_IRQDELAY_MEASURE                   0
#endif

#ifndef ESS_COMPLEX_ACCESSSPEED_MEASURE                 /* Set to 1 to let the app estimate ESC read/write speed at startup */
# define ESS_COMPLEX_ACCESSSPEED_MEASURE                0
#endif

/* The SLAVE_ defines below might be overwritten at runtime, see Init_GenericObjects() */

#ifndef SLAVE_SOFTWAREVERSION                           /* Your version string, used in CoE Object 0x100a and debug output */
# define SLAVE_SOFTWAREVERSION                          "esd Slave Sample (complex.c, built: " __DATE__ " " __TIME__ ")"
#endif

    /* Your ETG Vendor Id, used in CoE Object 0x1018.1 (Must match EEPROM value)
     *
     * TODO: !! Vendor Id has to be replaced !!
     *       Request Id here: http://www.ethercat.org/memberarea/vendor_id.asp
     *       Find existing Id here: http://www.ethercat.org/en/vendor_id_list.html
     */
#ifndef SLAVE_VENDORID
# define SLAVE_VENDORID                                 0xe0000017
#endif

    /*
     * Your device's product code, used in CoE Object 0x1018.2 (Must match EEPROM value).
     *
     * Product codes for esd hardware adapted in Init_GenericObjects()
     */
#ifndef SLAVE_PRODUCTCODE
# define SLAVE_PRODUCTCODE                              0x00000000
#endif

    /*
     * Your device's Revision Number, used in CoE Object 0x1018.3
     * (esd: 0x01000001 = Sample "Complex")
     */
#ifndef SLAVE_REVISIONNUMBER
# define SLAVE_REVISIONNUMBER                           0x01000001
#endif

    /*
     * Your device's Serial Number, used in CoE Object 0x1018.4
     * 0: Requested from HAL layer in Init_GenericObjects().
     */
#ifndef SLAVE_SERIALNUMBER
# define SLAVE_SERIALNUMBER                             0
#endif

#ifndef SLAVE_HARDWAREVERSION                           /* Your Slave's Hardware Version, used in CoE Object 0x1009 */
# define SLAVE_HARDWAREVERSION                          "N/A"
#endif

#ifndef SLAVE_DEVICENAME                                /* Your Slave's Device Name, used in CoE Object 0x1008 */
# define SLAVE_DEVICENAME                               "N/A"
#endif

#ifndef SLAVE_DEVICETYPE                                /* Your Slave's Device Type, used in CoE Object 0x1000 */
# define SLAVE_DEVICETYPE                               5001U /* 5001 = Modular Device, Profile No. = 0 (so this fits almost any slave that does not implement a specific profile) */
#endif


#ifndef SLAVE_MBX_IN_DEF                                /* Default size for Mailbox In */
# define SLAVE_MBX_IN_DEF                               522
#endif

#ifndef SLAVE_MBX_OUT_DEF                               /* Default size for Mailbox Out */
# define SLAVE_MBX_OUT_DEF                              522
#endif

#ifndef SLAVE_MBX_IN_MAX                                /* Max size for Mailbox In */
# define SLAVE_MBX_IN_MAX                               1024
#endif

#ifndef SLAVE_MBX_OUT_MAX                               /* Max size for Mailbox Out */
# define SLAVE_MBX_OUT_MAX                              1024
#endif

#if (SLAVE_MBX_OUT_MAX > CFG_ESS_MAX_MBX_LEN) || (SLAVE_MBX_IN_MAX > CFG_ESS_MAX_MBX_LEN)
# error "Mailbox max size exceeds CFG_ESS_MAX_MBX_LEN"
#endif

    /*
     * If the stack supports 'Requesting ID' mechanism via register 0x134 this
     * should be preferred over the 'Direct ID' mechanism
     */
#if (CFG_ESS_EXPLICIT_DEVICE_ID) && (ESS_COMPLEX_DIRECT_ID_ADO)
# error "Do not enable Direct ID mechanism if Requesting ID is available"
#endif

/* Other macros: */
/****************/
                                                        /* Used to get No. of elements in array */
#define ARRAY_LEN(arr)                                  ((uint32_t)(sizeof(arr) / sizeof((arr)[0])))
#define PRINT(x)                                        do { ESS_PRINTF("[Application] "); ESS_PRINTF x; ESS_PRINTF("\n"); } while (0)
#define MEM_DUMP(title, pData, len)                     do { uint32_t _currOffset = 0; ESS_PRINTF("%s\n0x%.4x: ", (title), _currOffset); if (pData) { while (_currOffset < (uint32_t)(len)) { ESS_PRINTF("%.2x ", ((uint8_t*)(pData))[_currOffset]); _currOffset++; if ((_currOffset % 16 == 0)) ESS_PRINTF("\n0x%.4x: ", _currOffset); else if ((_currOffset % 8) == 0) ESS_PRINTF(" "); } } ESS_PRINTF("\n"); } while (0)

#if (ESS_COMPLEX_VERBOSE)
# define PRINT_VERBOSE                                  PRINT
#else
# define PRINT_VERBOSE(x)
#endif

#define CHECK_RES()                                     CheckESSResult(res, __LINE__)

#if defined(_MSC_VER)
# pragma warning(disable:4127)                          /* avoid "conditional expression is constant" warnings... */
#endif



/* Declarations: */
/*****************/
static void cbInOutputsActivate(ESS_CBDATA_INOUTPUTS_ACTIVATE* cbData);
static void cbOutputsUpdated(ESS_CBDATA_SM_EVENT* cbData);
static void cbCoEReadWrite(ESS_CBDATA_COE_READWRITE* cbData);
static void cbCoEEvent(ESS_CBDATA_COE_EVENT* cbData);
static void cbDCEvent(ESS_CBDATA_DC* cbData);
static void cbCyclic(ESS_CBDATA_CYCLIC* cbData);
static void cbVoE(ESS_CBDATA_VOE* cbData);
static void cbSoE(ESS_CBDATA_SOE* cbData);
static void cbStateRequest(ESS_CBDATA_STATE_REQUEST* cbData);
#if (CFG_ESS_SUPPORT_FOE)
static void cbFoEOpen(ESS_CBDATA_FOE_OPEN* cbData);
static void cbFoEClose(ESS_CBDATA_FOE_CLOSE* cbData);
static void cbFoEData(ESS_CBDATA_FOE_DATA* cbData);
#endif
#if (CFG_ESS_EEPROM_EMULATION)
static void cbEEPROMEmulation(ESS_CBDATA_EEPROM_EMULATION* cbData);
#endif



/*
 * A few statistics that are recorded by the stack (Requires CFG_ESS_SERVE_STATISTICS=1).
 * Reported in CoE object 0x2110.
 */
static ESS_STATISTICS essStats;


/* Constants: */
/**************/

/* PDO contents: { {ObjIdx, SubIdx, Length}, ... } */
/* RxPDOs (Outputs): */
static const ESS_PDO_ENTRY PDO1600Entries[] = { ESS_MAP_ENTRY(0x2000, 0, 32) };
static const ESS_PDO_ENTRY PDO1601Entries[] = { ESS_MAP_ENTRY( 0x2001, 0, 32) };
/* TxPDOs (Inputs): */
static const ESS_PDO_ENTRY PDO1a00Entries[] = { ESS_MAP_ENTRY(0x2010, 0, 32) };
static const ESS_PDO_ENTRY PDO1a01Entries[] = { ESS_MAP_ENTRY(0x2011, 0, 32),
                                                ESS_MAP_ENTRY(0x2012, 0, 8 ) };

/* Default PDO assignment: (set in SetDefaultPDOs()) */
static const uint16_t SM2DefPDOs[] = { 0x1600 };
static const uint16_t SM3DefPDOs[] = { 0x1a00 };


/* SM configuration: (Assuming ESC memory 0x1000..0x2fff here: OK for ET1100, but with other ESCs with less than 8k RAM you might have to realign these sample SM areas) */
static const ESS_SM_CONFIGURATION smConfigs[] =
{
    { /* SM0: MBoxOut */
        46, /* minSize */
        SLAVE_MBX_OUT_DEF,
        SLAVE_MBX_OUT_MAX,
        0x1000, /* startAddr */
        SM_TYPE_MBXOUT | REG_MASK_SMCONTROL_PDIINT /* contrByte */
    },

    { /* SM1: MBoxIn */
        46, /* minSize */
        SLAVE_MBX_IN_DEF,
        SLAVE_MBX_IN_MAX,
        0x1400, /* startAddr */
        SM_TYPE_MBXIN | REG_MASK_SMCONTROL_PDIINT /* contrByte */
    },

    { /* SM2: Outputs */
        0, /* minSize */
        4, /* defSize */
        0, /* maxSize (0: unchecked) */
        0x1800, /* startAddr */
        SM_TYPE_OUTPUTS | REG_MASK_SMCONTROL_PDIINT /* contrByte */
    },

    { /* SM3: Inputs */
        0, /* minSize */
        4, /* defSize */
        0, /* maxSize (0: unchecked) */
        0x2400, /* startAddr */
        SM_TYPE_INPUTS | REG_MASK_SMCONTROL_PDIINT /* contrByte */
    }

};

static const ESS_CALLBACKS callbacks = {
    /* ESS_CB_CYCLIC */
    &cbCyclic,

    /* ESS_CB_STATE_REQUEST */
    &cbStateRequest,

    /* ESS_CB_SYNCMANAGER */
    NULL,

    /* ESS_CB_INOUTPUTS_ACTIVATE */
    &cbInOutputsActivate,

    /* ESS_CB_OUTPUTS_UPDATED */
    &cbOutputsUpdated,

    /* ESS_CB_INPUTS_UPDATED */
    NULL,

    /* ESS_CB_COE_READWRITE */
    &cbCoEReadWrite,

    /* ESS_CB_COE_EVENT */
    &cbCoEEvent,

#if (CFG_ESS_SUPPORT_FOE)
    /* ESS_CB_FOE_OPEN */
    &cbFoEOpen,
    /* ESS_CB_FOE_CLOSE */
    &cbFoEClose,
    /* ESS_CB_FOE_DATA */
    &cbFoEData,
#else
    NULL,
    NULL,
    NULL,
#endif

    /* ESS_CB_EOE_SETIPPARAM */
    NULL,

    /* ESS_CB_EOE_SETADDRFILTER */
    NULL,

    /* ESS_CB_EOE_FRAME */
    NULL,

    /* ESS_CB_VOE */
    &cbVoE,

    /* ESS_CB_EEPROM_EMULATION */
#if (CFG_ESS_EEPROM_EMULATION)
    &cbEEPROMEmulation,
#else
    NULL,
#endif

    /* ESS_CB_DC */
    &cbDCEvent,

    /* ESS_CB_SOE */
    &cbSoE,

    /* ESS_CB_AOE */
    NULL,
};

/* Needed to initialize the stack with essOpen(): */
static const ESS_CONFIGURATION config = {
    ESS_ABI_VERSION,

    /* Config flags */
    ESS_CONFIG_FLAGS_USE_ISR | ESS_CONFIG_FLAGS_APPSUPP_BOOTSTRAP, /* ESS_CONFIG_FLAGS_NONE */

    /* Timer interval for cyclic stack call / callback. In microseconds (Accuracy, etc. depends on hardware/platform. essStart() might also return ESS_RESULT_INVALID_TIMER_CONFIG when this value is too high or too low) */
    1 * 1000, /* us */

    /* Custom Tag, for later retrieval by essGetTag() */
    NULL,

    /* All stack callbacks */
    &callbacks,

    /* See smConfigs[] */
    smConfigs,
    ARRAY_LEN(smConfigs),
    0,

    /* Reserved, must be 0 */
    { 0, 0 },

    /* Statistics */
    &essStats,

    /* Reserved */
    NULL,
};



/* Variables: */
/**************/
    /* Our handle to the ESC. (Stack needs this e.g. to distinguish between different ESCs) */
static ESS_HANDLE hDev;

    /* Set in cbStateRequest(), to track current state */
static ESC_STATE currESCState = ESC_STATE_INIT;

    /* Device Identification Value according to ETG.1020. A value of 0 means invalid */
static uint16_t devIdValue = 0;

/* Input/Output variables (Must/will be stored as little endian) */
/*****************************************************************/
static uint32_t input1;
static uint32_t input2;
static uint8_t input3;
static uint32_t output1;
static uint32_t output2;

static const uint8_t testEntry1Min = 10;
static const uint8_t testEntry1Max = 100;
static const uint8_t testEntry1Def = 50;
static uint8_t testEntry1 = 50;

static const int8_t testEntry2Min = -50;
static const int8_t testEntry2Max = 50;
static const int8_t testEntry2Def = -20;
static int8_t testEntry2 = -20;

static const uint16_t testEntry3Min = 1000;
static const uint16_t testEntry3Max = 10000;
static const uint16_t testEntry3Def = 5000;
static uint16_t testEntry3 = 5000;

static const int16_t testEntry4Min = -5000;
static const int16_t testEntry4Max = 5000;
static const int16_t testEntry4Def = -2000;
static int16_t testEntry4 = -2000;

static const uint32_t testEntry5Min = 100000;
static const uint32_t testEntry5Max = 1000000;
static const uint32_t testEntry5Def = 500000;
static uint32_t testEntry5 = 500000;

static const int32_t testEntry6Min = -500000;
static const int32_t testEntry6Max = 500000;
static const int32_t testEntry6Def = -200000;
static int32_t testEntry6 = -200000;

/* Needed for our outputs watchdog */
static ESS_BOOL localOutputHandlerActive;
/* Checked in cbCyclic(), reset in cbOutputsUpdated() */
static ESS_TIMESTAMP lastOutputs;
static ESS_TIMESTAMP pdWatchdogValue;



#if (ESS_COMPLEX_JITTER_MEASURE)
# define JITTER_MEASURE_PRINTINTERVAL                   4000                    /* No. of cycles between debug outputs */
# define JITTER_MEASURE_RESET()                         JitterMeasureReset()    /* Called each time out/inputs are started/stopped */
# define JITTER_MEASURE_UPDATE()                        JitterMeasureUpdate()   /* Called each time new outputs were received */
#else
# define JITTER_MEASURE_RESET()                         do { } while (0)
# define JITTER_MEASURE_UPDATE()                        do { } while (0)
#endif

#if (ESS_COMPLEX_JITTER_MEASURE)

static uint32_t jmNumUpdates;

/* cycle: time between two output events */
static uint32_t jmMinCycle;
static uint32_t jmMaxCycle;
static uint64_t jmTotalCycles;
static uint32_t jmNumCycles;

/* delta: difference between last two output events */
static uint64_t jmTotalDelta;
static uint32_t jmNumDeltas;

/* Called each time out/inputs are started/stopped and after Jdebug output was shown (every JITTER_MEASURE_PRINTINTERVAL cycles) */
static void JitterMeasureReset(void)
{
    jmNumUpdates = 0;

    jmMinCycle = (uint32_t)-1;
    jmMaxCycle = 0;

    jmTotalDelta = 0;
    jmNumDeltas = 0;

    jmTotalCycles = 0;
    jmNumCycles = 0;
}

/* Called each time new outputs were received */
static void JitterMeasureUpdate(void)
{
    const uint32_t tsNow = essESCRead32(hDev, ESC_REG_DCSYSTEMTIME);
    static uint32_t lastTs;

    jmNumUpdates++;

    if (jmNumUpdates > 1) {
        static uint32_t lastCycle;
        const uint32_t cycle = tsNow - lastTs;

        jmTotalCycles += cycle;
        jmNumCycles++;

        if (jmNumUpdates > 2) {
            const uint32_t delta = (cycle > lastCycle) ? cycle - lastCycle : lastCycle - cycle;

            jmTotalDelta += delta;
            jmNumDeltas++;
        }

        if (cycle < jmMinCycle)
            jmMinCycle = cycle;

        if (cycle > jmMaxCycle)
            jmMaxCycle = cycle;

        lastCycle = cycle;
    }

    lastTs = tsNow;

    if (jmNumDeltas >= JITTER_MEASURE_PRINTINTERVAL) {
        const uint32_t avgCycle = (uint32_t)(jmTotalCycles / jmNumCycles);
        const uint32_t avgDelta = (uint32_t)(jmTotalDelta / jmNumDeltas);

        PRINT(("Cycle time: %u.%u us [%u..%u us], average delta: %u.%u us", avgCycle / 1000, (avgCycle % 1000) / 100, jmMinCycle / 1000, jmMaxCycle / 1000, avgDelta / 1000, (avgDelta % 1000) / 100));
        JitterMeasureReset();
    }
}

#endif /* ESS_COMPLEX_JITTER_MEASURE */

#if (ESS_COMPLEX_IRQDELAY_MEASURE)

/* Called at every ESS_DC_EVENT_SYNC0 */
static void SYNC0IRQDelayTest(ESS_HANDLE h)
{
    static int read_delay;
    static uint64_t next;
    volatile uint64_t now;
    int irq_delay;

    essESCRead64To(h, ESC_REG_DCSYSTEMTIME, &now);
    irq_delay = (int)(now - next);

    essESCRead64To(h, ESC_REG_SYNC0NEXT, &next);

#if 0 /* 0 to disable read delay calculation */
    if (read_delay == 0) {
        const int testcount = 1000;
        uint64_t rdsum = 0;
        int rdmin = 2000000000;
        int rdmax = -2000000000;
        int i;

        for (i = 0; i < testcount; ++i) {
            volatile uint64_t t0;
            volatile uint64_t t1;
            int rd;

            essESCRead64To(h, ESC_REG_DCSYSTEMTIME, &t0);
            essESCRead64To(h, ESC_REG_DCSYSTEMTIME, &t1);
            rd = (int)(t1 - t0);

            rdsum += rd;
            if (rd < rdmin)
                rdmin = rd;
            if (rd > rdmax)
                rdmax = rd;
        }

        read_delay = rdmin;
        PRINT(("SYNC0 IRQ Delay: minimum read overhead = %u ns [max: %u, avg: %u]", read_delay, rdmax, (int)(rdsum / testcount)));

        essESCRead64To(h, ESC_REG_SYNC0NEXT, &next);
        return;
    }
#endif

    irq_delay -= read_delay;
    /* Negatives values are e.g. possible when read_delay is better than calculated */
    if (irq_delay < 0)
        irq_delay = 0;

    PRINT(("SYNC0 IRQ Delay: %4d.%d us", irq_delay / 1000, (irq_delay % 1000) / 100));
}

#endif /* ESS_COMPLEX_IRQDELAY_MEASURE */

#if (ESS_COMPLEX_ACCESSSPEED_MEASURE)

/* Called only once (by Init()) */
static void ESCAccessSpeedMeasure(ESS_HANDLE h)
{
    uint32_t buf[128];
    const uint16_t len = (uint16_t)sizeof(buf);
    uint32_t dur;
    int i;

#if (CFG_HAL_NEEDS_EVENTS)
# error "ESCAccessSpeedMeasure() (ESS_COMPLEX_ACCESSSPEED_MEASURE) is probably not realistic with CFG_HAL_NEEDS_EVENTS!"
#endif

    for (i = 0; i < ARRAY_LEN(buf); ++i)
        buf[i] = i;

    dur = 2000000000;
    for (i = 0; i < 100; ++i) {
        const uint32_t start = essESCRead32(h, ESC_REG_DCSYSTEMTIME);
        uint32_t d;

        essESCRead(h, 0x1000, buf, len);
        d = essESCRead32(h, ESC_REG_DCSYSTEMTIME) - start;
        if (d < dur)
            dur = d;
    }
    PRINT(("ESC Access Speed: Read = %u ns/byte", dur / len));

    dur = 2000000000;
    for (i = 0; i < 100; ++i) {
        const uint32_t start = essESCRead32(h, ESC_REG_DCSYSTEMTIME);
        uint32_t d;

        essESCWrite(h, 0x1000, buf, len);
        d = essESCRead32(h, ESC_REG_DCSYSTEMTIME) - start;
        if (d < dur)
            dur = d;
    }
    PRINT(("ESC Access Speed: Write = %u ns/byte", dur / len));
}

#endif /* ESS_COMPLEX_ACCESSSPEED_MEASURE */







/* Returns 0 when res == ESS_RESULT_SUCCESS. Also debug prints error when failed etc. */
static void CheckESSResult(ESS_RESULT res, int lineNo)
{
    if (res != ESS_RESULT_SUCCESS) {
        PRINT(("ERROR: ess call in line %d failed with '%s'", lineNo, essFormatResult(res)));
    }
}

/* Called before starting slave and then every time slave goes back to "Init" state */
static void SetDefaultPDOs(void)
{
    ESS_RESULT res;

    /* Set PDO configuration: (When ESI describes objects too, source and ESI must match!) */
    res = essODUpdatePDOConfiguration(hDev, 0x1600, "RxPDO1", PDO1600Entries, ARRAY_LEN(PDO1600Entries), ESS_FALSE); CHECK_RES();
    res = essODUpdatePDOConfiguration(hDev, 0x1601, "RxPDO2", PDO1601Entries, ARRAY_LEN(PDO1601Entries), ESS_FALSE); CHECK_RES();
    res = essODUpdatePDOConfiguration(hDev, 0x1a00, "TxPDO1", PDO1a00Entries, ARRAY_LEN(PDO1a00Entries), ESS_FALSE); CHECK_RES();
    res = essODUpdatePDOConfiguration(hDev, 0x1a01, "TxPDO2", PDO1a01Entries, ARRAY_LEN(PDO1a01Entries), ESS_FALSE); CHECK_RES();

    /* Set default assignment: (Active when master does not upload a PDO assignment, matches ESI PDOs with given "Sm" attribute) */
    res = essODUpdatePDOAssignment(hDev, ESS_SM_2, SM2DefPDOs, ARRAY_LEN(SM2DefPDOs), ESS_TRUE); CHECK_RES();
    res = essODUpdatePDOAssignment(hDev, ESS_SM_3, SM3DefPDOs, ARRAY_LEN(SM3DefPDOs), ESS_TRUE); CHECK_RES();
}

/* Called by cbStateRequest() whenever state changes to Init */
static void StateChangedToInit(void)
{
    SetDefaultPDOs();

    /* Explicit Device ID via 'Direct ID' mechanism */
#if (ESS_COMPLEX_DIRECT_ID_ADO)
    *(uint8_t *)ESS_COMPLEX_DIRECT_ID_ADO       = (uint8_t)(devIdValue & 0xFF);
    *(uint8_t *)(ESS_COMPLEX_DIRECT_ID_ADO + 1) = (uint8_t)((devIdValue >> 8) & 0xFF);
#endif
}

/*
 * Setup the default CoE objects.
 * 
 * Overwrite defines in this code here before essODAddGenericObjects() is called.
 */
static void Init_GenericObjects(void)
{
    static uint32_t slaveProductCode = SWAP32_HOST2LE(SLAVE_PRODUCTCODE);
    static uint32_t slaveRevisionNo  = SWAP32_HOST2LE(SLAVE_REVISIONNUMBER);
    static uint32_t slaveDeviceType  = SWAP32_HOST2LE(SLAVE_DEVICETYPE);
    static uint32_t slaveVendorId    = SWAP32_HOST2LE(SLAVE_VENDORID);
    static uint32_t slaveSerial      = SWAP32_HOST2LE(SLAVE_SERIALNUMBER);
    const char* slaveDeviceName      = SLAVE_DEVICENAME;
    const char* slaveHWVersion       = SLAVE_HARDWAREVERSION;
    uint32_t buf32;
    ESS_RESULT res;

    /* If slave's HAL/driver supports a serial number we can overwrite our define with that value: */
    res = essIoctl(hDev, ESS_IOCTL_GET_DEV_SERIALNO32, &buf32, (uint32_t)sizeof(buf32));
    if (res == ESS_RESULT_SUCCESS) {
        slaveSerial = SWAP32_HOST2LE(buf32);
        PRINT_VERBOSE(("IOCTL detected SerialNo: slaveSerial changed to 0x%.8x", buf32));
    }

    /* For esd slaves we also try to adapt the product code and HW Version: */
    if (slaveVendorId == SWAP32_HOST2LE(0xe0000017)) {
        char s[32];
        uint32_t u;

        res = essIoctl(hDev, ESS_IOCTL_GET_DEV_ORDERNOSTR, s, (uint32_t)sizeof(s));
        if (res == ESS_RESULT_SUCCESS) {
            /* Match esd hardware product number "E.110x.0x" */
            if ((s[0] == 'E') && (s[1] == '.') && (s[2] == '1') && (s[3] == '1') &&
                (s[4] == '0') && (s[6] == '.') && (s[7] == '0') && (s[9] == '\0')) {
                switch(s[5])
                {
                case '0':
                    slaveProductCode = SWAP32_HOST2LE(0x00000005);
                    slaveDeviceName = "ECS-PCIe/1100";
                    break;
                case '2':
                    slaveProductCode = SWAP32_HOST2LE(0x0000000a);
                    slaveDeviceName = "ECS-XMC/FPGA";
                    break;
                case '4':
                    slaveProductCode = SWAP32_HOST2LE(0x0000000d);
                    slaveDeviceName  = "ECS-PMC/FPGA";
                    break;
                case '6':
                    slaveProductCode = SWAP32_HOST2LE(0x0000000e);
                    slaveDeviceName  = "ECS-PCIe/FPGA";
                    break;
                default:
                    PRINT_VERBOSE(("Unknown esd product code %s", s));
                    break;
                }
                if(slaveProductCode != 0) {
                    PRINT_VERBOSE(("ESD HW detected: (ProdCode/DeveName) 0x%.8x / \"%s\"",
                        slaveProductCode, slaveDeviceName));
                }
            }
        }

        res = essIoctl(hDev, ESS_IOCTL_GET_DEV_HWVERSION32, &u, (uint32_t)sizeof(u));
        if (res == ESS_RESULT_SUCCESS) {
            slaveHWVersion = (u == 0) ? "0" : (u == 1) ? "1" : (u == 2) ? "2" : (u == 3) ? "3" : \
                                              (u == 4) ? "4" : (u == 5) ? "5" : (u == 6) ? "6" : \
                                              (u == 7) ? "7" : (u == 8) ? "8" : ">8";
            PRINT_VERBOSE(("ESD HW detected: (HW Version) %u \"%s\"", u, slaveHWVersion));
        }
    }

    res = essODAddGenericObjects(hDev, &slaveDeviceType, &slaveVendorId, &slaveProductCode,
                                 &slaveRevisionNo, &slaveSerial, slaveDeviceName, slaveHWVersion,
                                 SLAVE_SOFTWAREVERSION); CHECK_RES();
}

/* Called before we're starting. Setting up the dictionary/variables here */
static void Init(void)
{
    ESS_RESULT res;

#if (ESS_COMPLEX_ACCESSSPEED_MEASURE)
    ESCAccessSpeedMeasure(hDev);
#endif

    res = essODCreate(hDev, ESS_OD_FLAGS_HANDLE_SM_TYPES); CHECK_RES();

    Init_GenericObjects();

#if (ESS_COMPLEX_ID_RELOAD_OBJECT)
    /* Object 0x10E0: Device Identification Reload Object according to ETG.1020 */
    {
        static const ESS_OD_OBJECT_INFOS objInfos = { "Device Identification Reload", COE_DATATYPE_INVALID, COE_CODE_RECORD };
        static const ESS_OD_ENTRY_INFOS entryInfos0 = { "SubIndex 000", NULL, NULL, NULL, 0, COE_DATATYPE_USINT };
        static const ESS_OD_ENTRY_INFOS entryInfos1 = { "Configured Station Alias Register", NULL, NULL, NULL, 0, COE_DATATYPE_UINT };
        static const ESS_OD_ENTRY_INFOS entryInfos2 = { "Write Configured Station Alias Persistent", NULL, NULL, NULL, 0, COE_DATATYPE_BOOL };
        static const ESS_OD_ENTRY_INFOS entryInfos3 = { "Reload ID-Selector Value", NULL, NULL, NULL, 0, COE_DATATYPE_UINT };

        res = essODObjectAdd(hDev, COE_OBJ_DEVID_RELOAD, ESS_OD_OBJECT_FLAGS_NONE, &objInfos); CHECK_RES();

        /* "SubIndex 000", max. sub index (Always the first item for COE_CODE_RECORD/COE_CODE_ARRAY) */
        res = essODEntryAdd(hDev, COE_OBJ_DEVID_RELOAD, 0x00, 8, (void*)"\x03", COE_ACCESS_R, ESS_OD_ENTRY_FLAGS_NONE, &entryInfos0); CHECK_RES();

        /* "Configured Station Alias Register": 16 bit unsigned, Read/Write, data aqcuired by callback. (cbCoEReadWrite()) */
        res = essODEntryAdd(hDev, COE_OBJ_DEVID_RELOAD, 0x01, 16, NULL, COE_ACCESS_RW, ESS_OD_ENTRY_FLAGS_NONE, &entryInfos1); CHECK_RES();

        /* "Write Configured Station Alias Persistent": 8 bit Read Only. Return constant  */
        res = essODEntryAdd(hDev, COE_OBJ_DEVID_RELOAD, 0x02, 1, (void*)"\x00", COE_ACCESS_R, ESS_OD_ENTRY_FLAGS_NONE, &entryInfos2); CHECK_RES();

        /* "Reload ID selector value": 16 bit unsigned, Read only */
        res = essODEntryAdd(hDev, COE_OBJ_DEVID_RELOAD, 0x03, 16, &devIdValue, COE_ACCESS_R, ESS_OD_ENTRY_FLAGS_NONE, &entryInfos3); CHECK_RES();
    }
#endif

    /* Object 0x2000: "Output1", RX-Mappable (Output, mapped by default), read/write */
    {
        static const ESS_OD_OBJECT_INFOS objInfos = { "Output1", COE_DATATYPE_UDINT, COE_CODE_VARIABLE };
        static const ESS_OD_ENTRY_INFOS entryInfos0 = { "Output1", NULL, NULL, NULL, 0, COE_DATATYPE_UDINT };

        res = essODObjectAdd(hDev, 0x2000, ESS_OD_OBJECT_FLAGS_NONE, &objInfos); CHECK_RES();
        res = essODEntryAdd(hDev, 0x2000, 0x00, 32, &output1, COE_ACCESS_RXMAPPABLE | COE_ACCESS_RW, ESS_OD_ENTRY_FLAGS_NONE, &entryInfos0); CHECK_RES();
    }
    /* Object 0x2001: "Output2", RX-Mappable (Output, mapped optional), read/write */
    {
        static const ESS_OD_OBJECT_INFOS objInfos = { "Output2", COE_DATATYPE_UDINT, COE_CODE_VARIABLE };
        static const ESS_OD_ENTRY_INFOS entryInfos0 = { "Output2", NULL, NULL, NULL, 0, COE_DATATYPE_UDINT };

        res = essODObjectAdd(hDev, 0x2001, ESS_OD_OBJECT_FLAGS_NONE, &objInfos); CHECK_RES();
        res = essODEntryAdd(hDev, 0x2001, 0x00, 32, &output2, COE_ACCESS_RXMAPPABLE | COE_ACCESS_RW, ESS_OD_ENTRY_FLAGS_NONE, &entryInfos0); CHECK_RES();
    }
    /* Object 0x2003: "Output3", not mappable, write only, data acquired by callback */
    {
        static const ESS_OD_OBJECT_INFOS objInfos = { "Output3", COE_DATATYPE_UDINT, COE_CODE_VARIABLE };
        static const ESS_OD_ENTRY_INFOS entryInfos0 = { "Output3", NULL, NULL, NULL, 0, COE_DATATYPE_UDINT };

        res = essODObjectAdd(hDev, 0x2003, ESS_OD_OBJECT_FLAGS_NONE, &objInfos); CHECK_RES();
        res = essODEntryAdd(hDev, 0x2003, 0x00, 32, NULL, COE_ACCESS_W, ESS_OD_ENTRY_FLAGS_NONE, &entryInfos0); CHECK_RES();
    }


    /* Object 0x2010: "Input1", TX-Mappable (Input, mapped by default), read/write */
    {
        static const ESS_OD_OBJECT_INFOS objInfos = { "Input1", COE_DATATYPE_UDINT, COE_CODE_VARIABLE };
        static const ESS_OD_ENTRY_INFOS entryInfos0 = { "Input1", NULL, NULL, NULL, 0, COE_DATATYPE_UDINT };

        res = essODObjectAdd(hDev, 0x2010, ESS_OD_OBJECT_FLAGS_NONE, &objInfos); CHECK_RES();
        res = essODEntryAdd(hDev, 0x2010, 0x00, 32, &input1, COE_ACCESS_TXMAPPABLE | COE_ACCESS_RW, ESS_OD_ENTRY_FLAGS_NONE, &entryInfos0); CHECK_RES();
    }
    /* Object 0x2011: "Input2", TX-Mappable (Input, mapped optional), read/write */
    {
        static const ESS_OD_OBJECT_INFOS objInfos = { "Input2", COE_DATATYPE_UDINT, COE_CODE_VARIABLE };
        static const ESS_OD_ENTRY_INFOS entryInfos0 = { "Input2", NULL, NULL, NULL, 0, COE_DATATYPE_UDINT };

        res = essODObjectAdd(hDev, 0x2011, ESS_OD_OBJECT_FLAGS_NONE, &objInfos); CHECK_RES();
        res = essODEntryAdd(hDev, 0x2011, 0x00, 32, &input2, COE_ACCESS_TXMAPPABLE | COE_ACCESS_RW, ESS_OD_ENTRY_FLAGS_NONE, &entryInfos0); CHECK_RES();
    }
    /* Object 0x2012: "Input3", TX-Mappable (Input, mapped optional), read/write */
    {
        static const ESS_OD_OBJECT_INFOS objInfos = { "Input3", COE_DATATYPE_BIT8, COE_CODE_VARIABLE };
        static const ESS_OD_ENTRY_INFOS entryInfos0 = { "Input3", NULL, NULL, NULL, 0, COE_DATATYPE_BIT8 };

        res = essODObjectAdd(hDev, 0x2012, ESS_OD_OBJECT_FLAGS_NONE, &objInfos); CHECK_RES();
        res = essODEntryAdd(hDev, 0x2012, 0x00, 8, &input3, COE_ACCESS_TXMAPPABLE | COE_ACCESS_RW, ESS_OD_ENTRY_FLAGS_NONE, &entryInfos0); CHECK_RES();
    }

    /* Object 0x2100: "TestObj", record with different entries */
    {
        static const ESS_OD_OBJECT_INFOS objInfos = { "Test Object", COE_DATATYPE_INVALID, COE_CODE_RECORD };
        static const ESS_OD_ENTRY_INFOS entryInfos0 = { "SubIndex 000", NULL, NULL, NULL, 0, COE_DATATYPE_USINT };
        static const ESS_OD_ENTRY_INFOS entryInfos1 = { "TickCount", NULL, NULL, NULL, 0, COE_DATATYPE_UDINT };
        static const ESS_OD_ENTRY_INFOS entryInfos2 = { "Test-String", NULL, NULL, NULL, 0, COE_DATATYPE_STRING };
        static const ESS_OD_ENTRY_INFOS entryInfos3 = { "Test-OctetString", NULL, NULL, NULL, 0, COE_DATATYPE_OCTETSTRING };
        static const ESS_OD_ENTRY_INFOS entryInfos4 = { "TestEntry1-u8", &testEntry1Min, &testEntry1Max, &testEntry1Def, 0, COE_DATATYPE_USINT };
        static const ESS_OD_ENTRY_INFOS entryInfos5 = { "TestEntry2-s8", (const uint8_t *)&testEntry2Min, (const uint8_t *)&testEntry2Max, (const uint8_t *)&testEntry2Def, 0, COE_DATATYPE_SINT };
        static const ESS_OD_ENTRY_INFOS entryInfos6 = { "TestEntry3-u16", (const uint8_t *)&testEntry3Min, (const uint8_t *)&testEntry3Max, (const uint8_t *)&testEntry3Def, 0, COE_DATATYPE_UINT };
        static const ESS_OD_ENTRY_INFOS entryInfos7 = { "TestEntry4-s16", (const uint8_t *)&testEntry4Min, (const uint8_t *)&testEntry4Max, (const uint8_t *)&testEntry4Def, 0, COE_DATATYPE_INT };
        static const ESS_OD_ENTRY_INFOS entryInfos8 = { "TestEntry5-u32", (const uint8_t *)&testEntry5Min, (const uint8_t *)&testEntry5Max, (const uint8_t *)&testEntry5Def, 0, COE_DATATYPE_UDINT };
	static const ESS_OD_ENTRY_INFOS entryInfos9 = { "TestEntry6-s32", (const uint8_t *)&testEntry6Min, (const uint8_t *)&testEntry6Max, (const uint8_t *)&testEntry6Def, 0, COE_DATATYPE_DINT };
        static const ESS_OD_ENTRY_INFOS entryInfos10 = { "Test-Emergency", NULL, NULL, NULL, 0, COE_DATATYPE_UINT };
        static const uint8_t maxSubIndex = 10;

        res = essODObjectAdd(hDev, 0x2100, ESS_OD_OBJECT_FLAGS_NONE, &objInfos); CHECK_RES();

        /* "SubIndex 000", max. sub index (Always the first item for COE_CODE_RECORD/COE_CODE_ARRAY) */
        res = essODEntryAdd(hDev, 0x2100, 0x00, 8, (void*)&maxSubIndex, COE_ACCESS_R, ESS_OD_ENTRY_FLAGS_NONE, &entryInfos0); CHECK_RES();

        /* "TickCount": 32 bit Read-only, data aqcuired by callback. (cbCoEReadWrite()) */
        res = essODEntryAdd(hDev, 0x2100, 0x01, 32, NULL, COE_ACCESS_R, ESS_OD_ENTRY_FLAGS_NONE, &entryInfos1); CHECK_RES();

        /* "Test-String": 512 byte string. data aqcuired by callback. (cbCoEReadWrite())  */
        res = essODEntryAdd(hDev, 0x2100, 0x02, 512 * 8, NULL, COE_ACCESS_RW, ESS_OD_ENTRY_FLAGS_NONE, &entryInfos2); CHECK_RES();

        /* "Test-String": 512 byte string. data aqcuired by callback. (cbCoEReadWrite())  */
        res = essODEntryAdd(hDev, 0x2100, 0x03, 512 * 8, NULL, COE_ACCESS_RW, ESS_OD_ENTRY_FLAGS_NONE, &entryInfos3); CHECK_RES();

        /* "TestEntry1": 8 bit unsigned, with min/max/default value  */
        res = essODEntryAdd(hDev, 0x2100, 0x04, 8, &testEntry1, COE_ACCESS_RW, ESS_OD_ENTRY_FLAGS_NONE, &entryInfos4); CHECK_RES();

        /* "TestEntry2": 8 bit signed, with min/max/default value  */
        res = essODEntryAdd(hDev, 0x2100, 0x05, 8, &testEntry2, COE_ACCESS_RW, ESS_OD_ENTRY_FLAGS_NONE, &entryInfos5); CHECK_RES();

        /* "TestEntry3": 16 bit unsigned, with min/max/default value  */
        res = essODEntryAdd(hDev, 0x2100, 0x06, 16, &testEntry3, COE_ACCESS_RW, ESS_OD_ENTRY_FLAGS_NONE, &entryInfos6); CHECK_RES();

        /* "TestEntry4": 16 bit signed, with min/max/default value  */
        res = essODEntryAdd(hDev, 0x2100, 0x07, 16, &testEntry4, COE_ACCESS_RW, ESS_OD_ENTRY_FLAGS_NONE, &entryInfos7); CHECK_RES();

        /* "TestEntry5": 32 bit unsigned, with min/max/default value  */
        res = essODEntryAdd(hDev, 0x2100, 0x08, 32, &testEntry5, COE_ACCESS_RW, ESS_OD_ENTRY_FLAGS_NONE, &entryInfos8); CHECK_RES();

        /* "TestEntry6": 32 bit signed, with min/max/default value  */
        res = essODEntryAdd(hDev, 0x2100, 0x09, 32, &testEntry6, COE_ACCESS_RW, ESS_OD_ENTRY_FLAGS_NONE, &entryInfos9); CHECK_RES();
        
        /* "Test-Emergency": 16 bit unsigned. data handled by callback. (cbCoEReadWrite())  */
        res = essODEntryAdd(hDev, 0x2100, 0x0A, 16, NULL, COE_ACCESS_W, ESS_OD_ENTRY_FLAGS_NONE, &entryInfos10); CHECK_RES();
    }


    /* Object 0x2110: "ess Statistics", record to show the essStats variable that is filled by the stack when CFG_ESS_SERVE_STATISTICS=1 */
    {
#if (!CFG_BIG_ENDIAN) /* Directly mapping essStats as object item data works only on little endian systems (stack just uses it as is, and EtherCAT data is little endian) */
        static const ESS_OD_OBJECT_INFOS objInfos = { "ess Statistics", COE_DATATYPE_INVALID, COE_CODE_RECORD };
        static const ESS_OD_ENTRY_INFOS entryInfos0 = { "SubIndex 000", NULL, NULL, NULL, 0, COE_DATATYPE_USINT };
        static const ESS_OD_ENTRY_INFOS entryInfos1 = { "SM0 Interrupts", NULL, NULL, NULL, 0, COE_DATATYPE_UDINT };
        static const ESS_OD_ENTRY_INFOS entryInfos2 = { "SM1 Interrupts", NULL, NULL, NULL, 0, COE_DATATYPE_UDINT };
        static const ESS_OD_ENTRY_INFOS entryInfos3 = { "SM2 Interrupts", NULL, NULL, NULL, 0, COE_DATATYPE_UDINT };
        static const ESS_OD_ENTRY_INFOS entryInfos4 = { "SM3 Interrupts", NULL, NULL, NULL, 0, COE_DATATYPE_UDINT };
        static const ESS_OD_ENTRY_INFOS entryInfos5 = { "SM4 Interrupts", NULL, NULL, NULL, 0, COE_DATATYPE_UDINT };
        static const ESS_OD_ENTRY_INFOS entryInfos6 = { "SM5 Interrupts", NULL, NULL, NULL, 0, COE_DATATYPE_UDINT };
        static const ESS_OD_ENTRY_INFOS entryInfos7 = { "SM6 Interrupts", NULL, NULL, NULL, 0, COE_DATATYPE_UDINT };
        static const ESS_OD_ENTRY_INFOS entryInfos8 = { "SM7 Interrupts", NULL, NULL, NULL, 0, COE_DATATYPE_UDINT };
        /* 16 SMs are supported by stack, latter 8 just ignored here... */
        static const ESS_OD_ENTRY_INFOS entryInfos11 = { "Total Stack calls", NULL, NULL, NULL, 0, COE_DATATYPE_UDINT };
        static const ESS_OD_ENTRY_INFOS entryInfos12 = { "Stack cyclic calls", NULL, NULL, NULL, 0, COE_DATATYPE_UDINT };
        static const uint8_t maxSubIndex = 0x12;

        res = essODObjectAdd(hDev, 0x2110, ESS_OD_OBJECT_FLAGS_NONE, &objInfos); CHECK_RES();

        /* "SubIndex 000", max. sub index (Always the first item for COE_CODE_RECORD/COE_CODE_ARRAY) */
        res = essODEntryAdd(hDev, 0x2110, 0x00, 8, (void*)&maxSubIndex, COE_ACCESS_R, ESS_OD_ENTRY_FLAGS_NONE, &entryInfos0); CHECK_RES();

        res = essODEntryAdd(hDev, 0x2110, 0x01, 32, &essStats.smInterrupts[0], COE_ACCESS_R, ESS_OD_ENTRY_FLAGS_NONE, &entryInfos1); CHECK_RES();
        res = essODEntryAdd(hDev, 0x2110, 0x02, 32, &essStats.smInterrupts[1], COE_ACCESS_R, ESS_OD_ENTRY_FLAGS_NONE, &entryInfos2); CHECK_RES();
        res = essODEntryAdd(hDev, 0x2110, 0x03, 32, &essStats.smInterrupts[2], COE_ACCESS_R, ESS_OD_ENTRY_FLAGS_NONE, &entryInfos3); CHECK_RES();
        res = essODEntryAdd(hDev, 0x2110, 0x04, 32, &essStats.smInterrupts[3], COE_ACCESS_R, ESS_OD_ENTRY_FLAGS_NONE, &entryInfos4); CHECK_RES();
        res = essODEntryAdd(hDev, 0x2110, 0x05, 32, &essStats.smInterrupts[4], COE_ACCESS_R, ESS_OD_ENTRY_FLAGS_NONE, &entryInfos5); CHECK_RES();
        res = essODEntryAdd(hDev, 0x2110, 0x06, 32, &essStats.smInterrupts[5], COE_ACCESS_R, ESS_OD_ENTRY_FLAGS_NONE, &entryInfos6); CHECK_RES();
        res = essODEntryAdd(hDev, 0x2110, 0x07, 32, &essStats.smInterrupts[6], COE_ACCESS_R, ESS_OD_ENTRY_FLAGS_NONE, &entryInfos7); CHECK_RES();
        res = essODEntryAdd(hDev, 0x2110, 0x08, 32, &essStats.smInterrupts[7], COE_ACCESS_R, ESS_OD_ENTRY_FLAGS_NONE, &entryInfos8); CHECK_RES();

        res = essODEntryAdd(hDev, 0x2110, 0x11, 32, &essStats.totalCalls, COE_ACCESS_R, ESS_OD_ENTRY_FLAGS_NONE, &entryInfos11); CHECK_RES();
        res = essODEntryAdd(hDev, 0x2110, 0x12, 32, &essStats.cyclicCalls, COE_ACCESS_R, ESS_OD_ENTRY_FLAGS_NONE, &entryInfos12); CHECK_RES();
#endif /* !CFG_BIG_ENDIAN */
    }

    StateChangedToInit();   /* Set default PDOs and Direct Device ID (if configured) */
}

static void InitALEventMask(void)
{
    uint32_t alEventMask = essESCRead32(hDev, ESC_REG_ALEVENTMASK);
    int i;

    /* Disable unwanted interrupts: */
    alEventMask &= ~0x00000040;  /* PD watchdog expired (Usually not needed) */
    for (i = 0; i < (int)ARRAY_LEN(smConfigs); ++i) {
        const ESS_SM_CONFIGURATION* const smConfig = &smConfigs[i];

        if ((smConfig->contrByte & 0x0f) == SM_TYPE_INPUTS)
            alEventMask &= ~(0x00000100 << i); /* Inputs Read (Callback not used in this sample. Additionally that interrupt is only cleared when
                                                  new inputs are written to SM buffer, i.e. it must be called in every cycle which might not be wanted) */
    }

    essESCWrite32(hDev, ESC_REG_ALEVENTMASK, alEventMask);
}

#if (ESS_COMPLEX_SHOW_PDO_CONFIG)
/* Displays all assigned PDOs and their configuration (For informational purposes only, e.g. when PDOs can be dynamically configured) */
static void ShowPDOAssignmentAndConfig(ESS_SM smIdx)
{
    uint16_t PDOs[CFG_ESS_COE_MAX_ITEMS_PER_PDOASSIGN_OBJECT];
    uint8_t numPDOsAssigned;
    ESS_RESULT res;

    numPDOsAssigned = ARRAY_LEN(PDOs);
    res = essODGetPDOAssignment(hDev, smIdx, &numPDOsAssigned, PDOs);
    if (res == ESS_RESULT_SUCCESS) {
        int i;
        PRINT(("%u PDOs assigned to SM%u", numPDOsAssigned, smIdx));

        for (i = 0; i < numPDOsAssigned; ++i) {
            ESS_PDO_ENTRY entries[CFG_ESS_COE_MAX_ITEMS_PER_PDO_OBJECT];
            uint8_t numEntries;
            int j;

            numEntries = ARRAY_LEN(entries);
            res = essODGetPDOConfiguration(hDev, PDOs[i], &numEntries, entries, 0);
            if (res != ESS_RESULT_SUCCESS)
                break;

            PRINT(("    %u entries in 0x%.4x", numEntries, PDOs[i]));
            for (j = 0; j < numEntries; ++j)
                PRINT(("        Obj 0x%.4x.%.2x, %u bit", entries[j].objIdx, entries[j].subIdx, entries[j].len));
        }
    }
}
#endif /* (ESS_COMPLEX_SHOW_PDO_CONFIG) */

/* Reads PD watchdog value as milliseconds value, 0 when disabled */
static uint32_t GetWatchdogValue(void)
{
    uint16_t multi = essESCRead16(hDev,ESC_REG_WATCHDOG_TIMEPROCESSDATA);
    uint16_t divider = essESCRead16(hDev, ESC_REG_WATCHDOG_DIVIDER);
    uint32_t baseNanoSecs;
    uint32_t us;

    baseNanoSecs = (divider + 2) * 40;
    if (baseNanoSecs > 65535) {
        baseNanoSecs /= 1000;
        us = multi * baseNanoSecs;
    } else {
        us = multi * baseNanoSecs;
        us /= 1000;
    }

    if (us == 0)
        return 0;

    us /= 1000;
    return (us > 0) ? us : 1;
}

/* Called for entries that were created with "data" set to "NULL" - stack reads/writes data for these entries by this callback */
static void cbCoEReadWrite(ESS_CBDATA_COE_READWRITE* cbData)
{
    PRINT_VERBOSE(("Stack wants to %s object 0x%.4x.%.2x", cbData->isRead ? "read data from" : "write data to", cbData->objIdx, cbData->entryIdx));

    if ((cbData->objIdx == 0x2003) && (cbData->entryIdx == 0) && (cbData->isRead == ESS_FALSE)) {
        uint32_t output3 = MAKE_DWORD(cbData->destSrc[0], cbData->destSrc[1], cbData->destSrc[2], cbData->destSrc[3]);
                           /* No cast here, cause CoE data is always little endian and we want to run this on big endian systems too, also we don't won't make assumptions about cbData->destSrc's alignment. */

        cbData->abortCode = COE_ABORTCODE_NONE;
        PRINT(("    output3 was written to 0x%.8x", output3));

        (void)output3;


    } else if (cbData->objIdx == 0x2100) {
        if ((cbData->entryIdx == 1) && (cbData->isRead) && (cbData->len == 4)) {
            uint32_t tickValue = essGetTime();

            /* Again: no cast here, to keep this code portable: */
            cbData->destSrc[0] = LO_BYTE(LO_WORD(tickValue));
            cbData->destSrc[1] = HI_BYTE(LO_WORD(tickValue));
            cbData->destSrc[2] = LO_BYTE(HI_WORD(tickValue));
            cbData->destSrc[3] = HI_BYTE(HI_WORD(tickValue));
            cbData->abortCode = COE_ABORTCODE_NONE;
            PRINT(("    Returned essGetTime() value as 0x2100.1's data"));

        } else if ((cbData->entryIdx == 2)||(cbData->entryIdx == 3)) {

            if (cbData->isRead) {
                static uint8_t value;
                uint32_t i;

                if (cbData->offset == 0) {
                    value = 48;
                }
                PRINT(("    Data for 0x2100.%d was read, offset = %u, len = %u", cbData->entryIdx, cbData->offset, cbData->len));

                /* So each data byte will the the segment number! */
                for (i = 0; i < cbData->len; ++i)
                    cbData->destSrc[i] = value;

                value++;
            } else {
                /* We'll just dump the data: */
                PRINT(("    Data for 0x2100.%d was written, offset = %u, len = %u", cbData->entryIdx, cbData->offset, cbData->len));
                MEM_DUMP("    Content:", cbData->destSrc, cbData->len);
            }

            cbData->abortCode = COE_ABORTCODE_NONE;
        } else if ((cbData->entryIdx == 10) && (!cbData->isRead) && (cbData->len == 2)) {
            ESS_COE_EMERGENCY e;
            ESS_RESULT res;
            e.errCode=0x1234;
            e.errReg=0x88;
            e.data[0]=cbData->destSrc[0];
            e.data[1]=cbData->destSrc[1];
            e.data[2]='e';
            e.data[3]='s';
            e.data[4]='d';
            res = essCoESendEmergency(cbData->hDev,&e);
            if(res == ESS_RESULT_SUCCESS) {
                PRINT(("    Test-Emergency sent"));
            } else {
                PRINT(("    Sending emergency failed with %d",res));
            }
        }
#if (ESS_COMPLEX_ID_RELOAD_OBJECT)
    } else if (cbData->objIdx == COE_OBJ_DEVID_RELOAD) {
        if ((cbData->entryIdx == 1) && (cbData->len == 2)) {
            uint16_t regStationAlias;
            if (cbData->isRead) {
                regStationAlias = essESCRead16(hDev, ESC_REG_STATIONALIAS);
                cbData->destSrc[0] = LO_BYTE(regStationAlias);
                cbData->destSrc[1] = HI_BYTE(regStationAlias);
            } else {
                regStationAlias = MAKE_WORD(cbData->destSrc[0], cbData->destSrc[1]);
                essESCWrite16(hDev, ESC_REG_STATIONALIAS, regStationAlias);
            }

            cbData->abortCode = COE_ABORTCODE_NONE;
        }
#endif
    }
}

/* Called whenever certain CoE events happen */
static void cbCoEEvent(ESS_CBDATA_COE_EVENT* cbData)
{
    (void)cbData;

    PRINT_VERBOSE(("+++ CoE event %s (%u) for object 0x%.4x.%.2x +++", ESS_OD_ENTRY_CALLBACK_TO_STRING(cbData->cbType), cbData->cbType, cbData->objIdx, cbData->entryIdx));
}

/* Called when Inputs/Outputs shall be started/stopped: (Even when no outputs exist, etc.) */
static void cbInOutputsActivate(ESS_CBDATA_INOUTPUTS_ACTIVATE* cbData)
{
    JITTER_MEASURE_RESET(); /* Only with ESS_COMPLEX_JITTER_MEASURE, else no effect */

    if (cbData->started) {

        if (cbData->inputs) {
            PRINT(("Start ESC In/Out process data update..."));

            /* Restart inputs: */
            {
                input3 = 0x05;
                input1 = SWAP32_HOST2LE(1000U);
                input2 = SWAP32_HOST2LE(100000U);

                PRINT(("*** Initialized input1/2/3 to %u/%u/%u", SWAP32_LE2HOST(input1), SWAP32_LE2HOST(input2), input3));
            }

        } else {
            PRINT(("Start local output update..."));

            pdWatchdogValue = GetWatchdogValue();

            localOutputHandlerActive = ESS_TRUE;
            lastOutputs = essGetTime();
        }

    } else {
        /* Inputs or outputs shall be stopped: */

        if (cbData->inputs) {
            PRINT(("Stop ESC In/Out process data update..."));
        } else {
            PRINT(("Stop local output update..."));
            localOutputHandlerActive = ESS_FALSE;
        }
    }
}

/* Called after output SM buffers were written by the master and stack copied that data to our variables (In Op state) */
static void cbOutputsUpdated(ESS_CBDATA_SM_EVENT* cbData)
{
    static uint32_t output1Prev;
    static uint32_t output2Prev;
    (void)cbData;

    JITTER_MEASURE_UPDATE(); /* Only with ESS_COMPLEX_JITTER_MEASURE, else no effect */

    /* If you want to track all changes to a variable by callbacks, don't forget that it might be changed by a CoE download, too. */
    if(cbData->sm & ESS_SM_EVENT_FLAG_SAFE_OUTPUTS) {
        /* PRINT(("*** Outputs in 'safe state'")); */
    } else {
        if (output1 != output1Prev) {
            PRINT(("*** output1 changed to %u", SWAP32_LE2HOST(output1)));
            output1Prev = output1;
        }

        if (output2 != output2Prev) {
            PRINT(("*** output2 changed to %u", SWAP32_LE2HOST(output2)));
            output2Prev = output2;
        }
    }

    lastOutputs = essGetTime();
}

/* Called by cbCyclic when in OP or SAFEOP */
static void updateInputs(void)
{
    static uint32_t nextUpdate;

    if (essGetTime() >= nextUpdate) {
        input1 = SWAP32_HOST2LE(SWAP32_LE2HOST(input1) + 1);
        input2 = SWAP32_HOST2LE(SWAP32_LE2HOST(input2) + 10);
        input3 = LO_BYTE((input3) ? input3 << 1 : 5);

        PRINT_VERBOSE(("*** Set input1/2/3 to %u/%u/%u", SWAP32_LE2HOST(input1), SWAP32_LE2HOST(input2), input3));

        /* To let the stack copy the entry data to the SM buffer: */
        essSyncInputs(hDev, ESS_ISYN_FLAGS_NONE);

        nextUpdate = essGetTime() + 1000;
    }
}

/* Called when a VoE packet was received (also needs CFG_ESS_SUPPORT_VOE=1) */
static void cbVoE(ESS_CBDATA_VOE* cbData)
{
    /* Header as currently defined in ETG.1020: */
    const uint32_t vendorId = MBX_VOE_GET_VENDID(cbData->mb);
    const uint16_t protocolId = MBX_VOE_GET_PROTID(cbData->mb);
    uint8_t* const protocolData = &cbData->mb->data[MBX_VOE_HEADERLEN];
    const int protocolDataLen = cbData->mb->length - MBX_VOE_HEADERLEN;

    /* Make sure VoE support is also enabled in the slave's ESI (.xml file and EEPROM) */

    (void)vendorId;
    (void)protocolId;
    (void)protocolData;
    (void)protocolDataLen;

    /* To answer the packet modify the data at cbData->mb->data etc. and set cbData->result to MBX_ERR_SUCCESS */
    PRINT(("Received VoE packet: Vendor: %u, protocol: %u, protDataLen: %u", vendorId, protocolId, protocolDataLen));
}

/* Called when an SoE packet was received (also needs CFG_ESS_SUPPORT_SOE=1) */
static void cbSoE(ESS_CBDATA_SOE* cbData)
{
    const uint8_t opCode = MBX_SOE_GET_OPCODE(cbData->mb);
    const uint16_t dataLen = MBX_SOE_GET_DATALEN(cbData->mb);
    /* ... */

    /* Make sure SoE support is also enabled in the slave's ESI (.xml file and EEPROM) */

    (void)opCode;
    (void)dataLen;


    /* To answer the packet modify the data at cbData->mb->data etc. and set cbData->result to MBX_ERR_SUCCESS */
    PRINT(("Received SoE packet: OpCode = %u, data len = %u", opCode, dataLen));
}

/* Called when a DC event occurred (SYNC unit and mapping must be enabled in EEPROM etc.) */
static void cbDCEvent(ESS_CBDATA_DC* cbData)
{
    switch (cbData->evtType)
    {
        /* Assuming ACK mode (SYNC pulse len = 0) for this sample: (and very high intervals so debug output doesn't hurt...) */
        case ESS_DC_EVENT_SYNC0:
#if (ESS_COMPLEX_IRQDELAY_MEASURE)
            SYNC0IRQDelayTest(hDev);
#else
            PRINT_VERBOSE(("SYNC0 Interrupt!"));
#endif
            break;

        case ESS_DC_EVENT_SYNC1:
            PRINT_VERBOSE(("SYNC1 Interrupt!"));
            break;

        case ESS_DC_EVENT_LATCH:
            PRINT_VERBOSE(("DC Latch Interrupt!"));
            break;

        default:
            break;
    }
}

/* Called cyclically by stack (after essStart() was called) */
static void cbCyclic(ESS_CBDATA_CYCLIC* cbData)
{
    (void)cbData;

    if ((currESCState == ESC_STATE_SAFEOP) || (currESCState == ESC_STATE_OP))
        updateInputs();

#if (ESS_COMPLEX_PD_WATCHDOG)
    if ((localOutputHandlerActive) && (pdWatchdogValue)) {
        ESS_TIMESTAMP t = essGetTime() - lastOutputs;

        if (t > pdWatchdogValue) {
            PRINT(("Did not receive outputs for %lu ms (max is %lu ms), going to safeop...",
                (unsigned long)t, (unsigned long)pdWatchdogValue));

            essSetLEDState(hDev, ESC_LED_TYPE_ERR, ESC_LED_STATE_FLASH_2); /* As specified in ETG1300 "Indicator and Labeling" */
            essIndicateError(hDev, ESC_STATE_SAFEOP, REG_VAL_ALSTATUSCODE_SMWATCHDOG); /* With CFG_ESS_SERVE_ERR_LED a state change would lead to ERR LED flashing, too. But as specified, the first LED state shall remain */
        }
    }
#else
    /* Avoid warnings about unused variables: */
    (void)localOutputHandlerActive;
    (void)pdWatchdogValue;
    (void)lastOutputs;
#endif
}


#if (CFG_ESS_SUPPORT_FOE)

    /* All FoE transfers in this sample are denied when password differs from this */
static const uint32_t foePassword = 0;

    /* Upload file size is set to random value in this sample */
static uint32_t foeUploadFileSize;

    /* Used to show checksum of up/downloaded files */
static uint32_t foeCRC;

    /* Simulation of an increasing number of busy replies */
static uint32_t foeSimulateBusy;
static uint32_t foeSimulateBusyCount;

/* From: "CRC-32 version 2.0.0 by Craig Bruce, 2006-04-29." (Public domain code) */
static unsigned long CRC32(unsigned long inCrc32, const void *buf, size_t bufLen)
{
    static const unsigned long crcTable[256] = { 0x00000000, 0x77073096, 0xEE0E612C, 0x990951BA, 0x076DC419, 0x706AF48F, 0xE963A535, 0x9E6495A3, 0x0EDB8832, 0x79DCB8A4, 0xE0D5E91E, 0x97D2D988, 0x09B64C2B, 0x7EB17CBD, 0xE7B82D07, 0x90BF1D91, 0x1DB71064, 0x6AB020F2, 0xF3B97148, 0x84BE41DE, 0x1ADAD47D,
            0x6DDDE4EB, 0xF4D4B551, 0x83D385C7, 0x136C9856, 0x646BA8C0, 0xFD62F97A, 0x8A65C9EC, 0x14015C4F, 0x63066CD9, 0xFA0F3D63, 0x8D080DF5, 0x3B6E20C8, 0x4C69105E, 0xD56041E4, 0xA2677172, 0x3C03E4D1, 0x4B04D447, 0xD20D85FD, 0xA50AB56B, 0x35B5A8FA, 0x42B2986C, 0xDBBBC9D6, 0xACBCF940, 0x32D86CE3,
            0x45DF5C75, 0xDCD60DCF, 0xABD13D59, 0x26D930AC, 0x51DE003A, 0xC8D75180, 0xBFD06116, 0x21B4F4B5, 0x56B3C423, 0xCFBA9599, 0xB8BDA50F, 0x2802B89E, 0x5F058808, 0xC60CD9B2, 0xB10BE924, 0x2F6F7C87, 0x58684C11, 0xC1611DAB, 0xB6662D3D, 0x76DC4190, 0x01DB7106, 0x98D220BC, 0xEFD5102A, 0x71B18589,
            0x06B6B51F, 0x9FBFE4A5, 0xE8B8D433, 0x7807C9A2, 0x0F00F934, 0x9609A88E, 0xE10E9818, 0x7F6A0DBB, 0x086D3D2D, 0x91646C97, 0xE6635C01, 0x6B6B51F4, 0x1C6C6162, 0x856530D8, 0xF262004E, 0x6C0695ED, 0x1B01A57B, 0x8208F4C1, 0xF50FC457, 0x65B0D9C6, 0x12B7E950, 0x8BBEB8EA, 0xFCB9887C, 0x62DD1DDF,
            0x15DA2D49, 0x8CD37CF3, 0xFBD44C65, 0x4DB26158, 0x3AB551CE, 0xA3BC0074, 0xD4BB30E2, 0x4ADFA541, 0x3DD895D7, 0xA4D1C46D, 0xD3D6F4FB, 0x4369E96A, 0x346ED9FC, 0xAD678846, 0xDA60B8D0, 0x44042D73, 0x33031DE5, 0xAA0A4C5F, 0xDD0D7CC9, 0x5005713C, 0x270241AA, 0xBE0B1010, 0xC90C2086, 0x5768B525,
            0x206F85B3, 0xB966D409, 0xCE61E49F, 0x5EDEF90E, 0x29D9C998, 0xB0D09822, 0xC7D7A8B4, 0x59B33D17, 0x2EB40D81, 0xB7BD5C3B, 0xC0BA6CAD, 0xEDB88320, 0x9ABFB3B6, 0x03B6E20C, 0x74B1D29A, 0xEAD54739, 0x9DD277AF, 0x04DB2615, 0x73DC1683, 0xE3630B12, 0x94643B84, 0x0D6D6A3E, 0x7A6A5AA8, 0xE40ECF0B,
            0x9309FF9D, 0x0A00AE27, 0x7D079EB1, 0xF00F9344, 0x8708A3D2, 0x1E01F268, 0x6906C2FE, 0xF762575D, 0x806567CB, 0x196C3671, 0x6E6B06E7, 0xFED41B76, 0x89D32BE0, 0x10DA7A5A, 0x67DD4ACC, 0xF9B9DF6F, 0x8EBEEFF9, 0x17B7BE43, 0x60B08ED5, 0xD6D6A3E8, 0xA1D1937E, 0x38D8C2C4, 0x4FDFF252, 0xD1BB67F1,
            0xA6BC5767, 0x3FB506DD, 0x48B2364B, 0xD80D2BDA, 0xAF0A1B4C, 0x36034AF6, 0x41047A60, 0xDF60EFC3, 0xA867DF55, 0x316E8EEF, 0x4669BE79, 0xCB61B38C, 0xBC66831A, 0x256FD2A0, 0x5268E236, 0xCC0C7795, 0xBB0B4703, 0x220216B9, 0x5505262F, 0xC5BA3BBE, 0xB2BD0B28, 0x2BB45A92, 0x5CB36A04, 0xC2D7FFA7,
            0xB5D0CF31, 0x2CD99E8B, 0x5BDEAE1D, 0x9B64C2B0, 0xEC63F226, 0x756AA39C, 0x026D930A, 0x9C0906A9, 0xEB0E363F, 0x72076785, 0x05005713, 0x95BF4A82, 0xE2B87A14, 0x7BB12BAE, 0x0CB61B38, 0x92D28E9B, 0xE5D5BE0D, 0x7CDCEFB7, 0x0BDBDF21, 0x86D3D2D4, 0xF1D4E242, 0x68DDB3F8, 0x1FDA836E, 0x81BE16CD,
            0xF6B9265B, 0x6FB077E1, 0x18B74777, 0x88085AE6, 0xFF0F6A70, 0x66063BCA, 0x11010B5C, 0x8F659EFF, 0xF862AE69, 0x616BFFD3, 0x166CCF45, 0xA00AE278, 0xD70DD2EE, 0x4E048354, 0x3903B3C2, 0xA7672661, 0xD06016F7, 0x4969474D, 0x3E6E77DB, 0xAED16A4A, 0xD9D65ADC, 0x40DF0B66, 0x37D83BF0, 0xA9BCAE53,
            0xDEBB9EC5, 0x47B2CF7F, 0x30B5FFE9, 0xBDBDF21C, 0xCABAC28A, 0x53B39330, 0x24B4A3A6, 0xBAD03605, 0xCDD70693, 0x54DE5729, 0x23D967BF, 0xB3667A2E, 0xC4614AB8, 0x5D681B02, 0x2A6F2B94, 0xB40BBE37, 0xC30C8EA1, 0x5A05DF1B, 0x2D02EF8D };
    unsigned char* byteBuf;
    unsigned long crc32;
    size_t i;

    crc32 = inCrc32 ^ 0xFFFFFFFF;
    byteBuf = (unsigned char*)buf;

    for (i = 0; i < bufLen; i++)
        crc32 = (crc32 >> 8) ^ crcTable[(crc32 ^ byteBuf[i]) & 0xFF];

    return (crc32 ^ 0xFFFFFFFF);
}


/* Called when an FoE transfer shall be started: */
static void cbFoEOpen(ESS_CBDATA_FOE_OPEN* cbData)
{
    PRINT(("FoE: Open file '%s' for %s",
        cbData->fileName, cbData->isUpload ? "READING" : "WRITING"));
    PRINT(("FoE:   P/W = 0x%.8x, Net mbx size: %d bytes",
        cbData->password, cbData->maxDataLen));

    /* Check password, access rights, etc, here: */
    if (cbData->password != foePassword) {
        cbData->errCode = FOE_ERRCODE_ACCESSDENIED;
        PRINT(("Denied FoE transfer due to invalid password!"));
        return;
    }

    if((cbData->fileNameLen >=4) && (strstr(cbData->fileName, "busy") != NULL)) {
        foeSimulateBusy = 1;
    } else {
        foeSimulateBusy = 0;
    }
    foeSimulateBusyCount = 0;

    if (cbData->isUpload) {
        /* Open file for reading: */

        /* This sample sends random data of random len: */
        if ((rand() & 0x3) == 0) /* ...with preference for lengths that are multiples of FoE transfer chunk len (if default SM len is used) */
            foeUploadFileSize = cbData->maxDataLen * ((rand() % 10) + 1);
        else
            foeUploadFileSize = (rand() & 0xfff) + 1;
        foeCRC = 0;
        PRINT(("FoE Upload: Starting: file size = %u", foeUploadFileSize));
        /* ... */

        cbData->errCode = FOE_ERRCODE_NONE;

    } else {
        /* Open file for writing: */

        /* This sample just calcs the CRC32 of the received file */
        foeCRC = 0;
        PRINT(("FoE Download: Starting"));
        /* ... */

        cbData->errCode = FOE_ERRCODE_NONE;
    }
}

/* Called when an FoE transfer was finished or aborted: */
static void cbFoEClose(ESS_CBDATA_FOE_CLOSE* cbData)
{
    if (cbData->unexpected)
        PRINT(("WARNING: FoE transfer incomplete!"));

    if (cbData->isUpload == ESS_FALSE) {
        /* Close received file here: */

        /* Sample handling just shows CRC: */
        if (!cbData->unexpected) {
            PRINT(("FoE Download: Finished, CRC32 = 0x%.8x", foeCRC));
            /* ... */
        }

    } else {
        /* Close sent file here: */

        /* Sample handling just shows CRC: */
        if (!cbData->unexpected) {
            PRINT(("FoE Upload: Finished, CRC32 = 0x%.8x", foeCRC));
            /* ... */
        }

    }
}

/*
 * Called every time when the next segment of data of an FoE transfer is received
 *  or has to be sent.
 * 
 * The cbData->errCode is replied to the server and has to be set accordingly:
 *   FOE_ERRCODE_NONE - Data successfully uploaded or downloaded
 *   FOE_ERRCODE_BUSY - Stall the FoE transfer with a busy reply
 *   FOE_ERRCODE_XXX  - Terminate the transfer
 *
 * Upload:   The cbData->len is set to cbData->maxDataLen by the stack and has to
 *           be set here to a smaller value to indicate the last segment.
 * Download: The cbData->len is set to a value smaller than cbData->maxDataLen
 *           by the stack to indicate the last segment.
 */
static void cbFoEData(ESS_CBDATA_FOE_DATA* cbData)
{
    /* This code simulates busy replies */
    if(foeSimulateBusy != 0) {
        if(foeSimulateBusyCount < foeSimulateBusy) {
            /* uint8_t done = (uint8_t)(((foeSimulateBusyCount + 1) * 100) / foeSimulateBusy); */
            uint8_t done = (uint8_t)(((foeSimulateBusyCount) * 100) / foeSimulateBusy);
            cbData->errCode = FOE_ERRCODE_BUSY(done);
            if (cbData->isUpload) {
                PRINT(("FoE Upload: Stall (%d %% ready) (%u byte at offset %u)",
                    done, cbData->len, cbData->offset));
            } else {
                PRINT(("FoE Download: Stall (%d %% ready) (%u byte at offset %u)",
                    done, cbData->len, cbData->offset));
            }
            foeSimulateBusyCount++;
            return;
        } else {
            foeSimulateBusy++;
            foeSimulateBusyCount = 0;
        }
    }
    
    if (cbData->isUpload) {
        /* Upload: Copy data from file to stack: */

        /*
         * Copy cbData->maxDataLen from file at offset cbData->offset to cbData->destSrc.
         * Set cbData->len either to cbData->maxDataLen or to a smaller number to
         * indicate to the stack that this is the last segment of the file.
         * Note: If the last segment's size is cbData->maxDataLen you have to set
         *       cbData->len with the next call to 0.
         */
        PRINT(("FoE Upload: Sending data chunk (offset %u)", cbData->offset));

        /* This sample just uploads random data: */
        {
            const int remainFileLen = foeUploadFileSize - cbData->offset;
            int i;

            if (remainFileLen > 0) {
                if (remainFileLen < cbData->len) {
                    cbData->len = (uint16_t)remainFileLen;  /* Last segment indication */
                }

                for (i = 0; i < cbData->len; ++i) {
                    cbData->destSrc[i] = (uint8_t)rand();
                }

                foeCRC = CRC32(foeCRC, cbData->destSrc, cbData->len);

            } else {
                cbData->len = 0;    /* Last segment indication */
            }

            PRINT(("FoE Upload: %u bytes uploaded", cbData->len));
        }
        /* ... */


        cbData->errCode = FOE_ERRCODE_NONE;

    } else {
        /* Download: Write data from stack to file: */

        /*
         * Copy cbData->len bytes from cbData->destSrc to the file at offset cbData->offset
         * The cbData->len is either cbData->maxDataLen or a smaller number (even 0) to
         * indicate to the application that this is the last segment of the file.
         */
        PRINT(("FoE Download: Received data chunk (%u bytes at offset %u)", cbData->len, cbData->offset));

        /* This sample does nothing with the data but updating CRC */
        foeCRC = CRC32(foeCRC, cbData->destSrc, cbData->len);
        /* ... */

        cbData->errCode = FOE_ERRCODE_NONE;
    }
}
#endif


#if (CFG_ESS_EEPROM_EMULATION)

    /*
     * If EEPROM_FILE_NAME is defined the virtualized EEPROM data is taken
     * from this file (which size has to match the size of eepromData[] below).
     * Undef EEPROM filename to initialize virtualized EEPROM data individually
     * or to provide static data.
     * 
     * Note: For the esd IP core based interfaces respective binary images are
     *       shipped together with the rspective ESI files. Individual EEPROM
     *       images can be created with the EtherCAT workbench.
     */
#define EEPROM_FILE_NAME           "eeprom.bin"

#ifdef EEPROM_FILE_NAME
static uint8_t eepromData[4096];
#else
static uint8_t eepromData[4096] = {
    /* Shall be initialized with data from your non-volatile storage before calling essOpen() */
};
#endif

static void cbEEPROMEmulation(ESS_CBDATA_EEPROM_EMULATION* cbData)
{
    if ((cbData->offset + cbData->len) > sizeof(eepromData)) {
        cbData->result = 1;
        return;
    }

    if (cbData->isRead) {
        memcpy(cbData->destSrc, &eepromData[cbData->offset], cbData->len);

    } else {
        memcpy(&eepromData[cbData->offset], cbData->destSrc, cbData->len);
        /* Now mark EEPROM as changed and flush it to non-volatile storage some time after last change */
    }

    cbData->result = 0;
}
#endif /* (CFG_ESS_EEPROM_EMULATION) */


/* Called when AL EVENT signaled an AL CONTROL event

   cbData->statusCode shows errors detected by stack.
   cbData->newState shows the state that will be set

   Application can override cbData->statusCode and cbData->newState
   or set cbData->tryAgain to 1 if it can't be decided yet (callback is
   called again then)
*/
static void cbStateRequest(ESS_CBDATA_STATE_REQUEST* cbData)
{
    const ESS_BOOL stackFoundNoError = (cbData->statusCode == REG_VAL_ALSTATUSCODE_NOERROR);

    PRINT(("---> Transition from %s (0x%.2x) to %s (0x%.2x):", ESC_STATE_TO_STRING(LO_BYTE(cbData->transition)), LO_BYTE(cbData->transition), ESC_STATE_TO_STRING(HI_BYTE(cbData->transition)), HI_BYTE(cbData->transition)));

    if (!stackFoundNoError) {
        PRINT(("     Error %s (0x%.4x)", REG_VAL_ALSTATUSCODE_TO_STRING(cbData->statusCode), cbData->statusCode));

        /* When stack already found an error we MUST NOT override cbData->statusCode in this callback */
    }

    switch (cbData->transition) {
        case ESC_TRANSITION_IB:
            if (stackFoundNoError) {
                /* Called only with ESS_CONFIG_FLAGS_APPSUPP_BOOTSTRAP */
                PRINT(("Entering BootStrap mode!"));
            }
            break;

        case ESC_TRANSITION_BI:
            break;

        case ESC_TRANSITION_IP:
            break;

        case ESC_TRANSITION_PS:
#if (ESS_COMPLEX_SHOW_PDO_CONFIG)
            if (stackFoundNoError)
                ShowPDOAssignmentAndConfig(ESS_SM_3);
#endif
            break;

        case ESC_TRANSITION_SO:
#if (ESS_COMPLEX_SHOW_PDO_CONFIG)
            if (stackFoundNoError)
                ShowPDOAssignmentAndConfig(ESS_SM_2);
#endif
            break;

        case ESC_TRANSITION_OS:
            break;

        case ESC_TRANSITION_OP:
            break;

        case ESC_TRANSITION_OI:
            break;

        case ESC_TRANSITION_SP:
            break;

        case ESC_TRANSITION_SI:
            break;

        case ESC_TRANSITION_PI:
            break;

        case ESC_TRANSITION_BB:
            break;

        case ESC_TRANSITION_II:
            break;

        case ESC_TRANSITION_PP:
            break;

        case ESC_TRANSITION_SS:
            break;

        case ESC_TRANSITION_OO:
            break;

        default:
            PRINT(("Unknown transition!, from %s (0x%.2x) to %s (0x%.2x)", ESC_STATE_TO_STRING(LO_BYTE(cbData->transition)), LO_BYTE(cbData->transition), ESC_STATE_TO_STRING(HI_BYTE(cbData->transition)), HI_BYTE(cbData->transition)));
            if (stackFoundNoError) {
                /* Change state from Op to SafeOp, else stay in current state */
                cbData->newState = (LO_BYTE(cbData->transition) == ESC_STATE_OP) ? ESC_STATE_SAFEOP : LO_BYTE(cbData->transition);
                cbData->statusCode = (ESC_STATE_KNOWN(HI_BYTE(cbData->transition))) ? REG_VAL_ALSTATUSCODE_INVALIDALCONTROL : REG_VAL_ALSTATUSCODE_UNKNOWNALCONTROL;
            }
            break;
    }

    PRINT(("---> New state will be %s (0x%.2x)", ESC_STATE_TO_STRING(cbData->newState), cbData->newState));

    currESCState = cbData->newState;

    if (currESCState == ESC_STATE_INIT)
        StateChangedToInit();

#if (CFG_ESS_EXPLICIT_DEVICE_ID)
    if(ESS_TRUE == cbData->devIdReq) {
        cbData->devId = devIdValue;
        PRINT(("---> Replied DevID: (0x%.4x)", cbData->devId));
    }
#endif

}




/* First function called by main() */
static void InitApplication(void)
{
#if defined(ESS_PLATFORM_LINUX)
    struct sched_param p;

    p.sched_priority = 90; /* Just a sample for common Linux systems, adapt it to your needs. */
    PRINT(("Entering SCHED_FIFO mode... with priority %d", p.sched_priority));

    if (sched_setscheduler(0, SCHED_FIFO, &p) == 0) {
        PRINT(("New scheduler policy is SCHED_FIFO, priority is %d", p.sched_priority));
    } else {
        PRINT(("Failed!"));
    }

#elif defined(ESS_PLATFORM_WINDOWS)
    PRINT(("Increasing Windows process priority..."));

    if (SetPriorityClass(GetCurrentProcess(), HIGH_PRIORITY_CLASS)) {
        /* SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_HIGHEST); */
    } else {
        PRINT(("Failed!"));
    }

#else
    /* ... */
#endif
}

/* Second function called by main() */
static void ShowVersionInfos(void)
{
    ESS_DEVICE_INDEX i;
    ESS_HANDLE h;

    PRINT(("App  : %s", SLAVE_SOFTWAREVERSION));
    PRINT(("Stack: %s", essGetVersion()));

    PRINT(("Trying to open all devices... (CFG_ESS_MAX_DEVICES = %u)", CFG_ESS_MAX_DEVICES));
    for (i = 0; i < CFG_ESS_MAX_DEVICES; ++i) {
        ESS_RESULT res;
        char bufstr[200];
        uint32_t buf32;

        res = essOpen(i, NULL, &h);
        if (res != ESS_RESULT_SUCCESS)
            continue;

        /* All these IOCTLs are not necessarily supported, so no error if there are no infos: */

        res = essIoctl(h, ESS_IOCTL_GET_HAL_BUILDINFOSTR, bufstr, (uint32_t)sizeof(bufstr));
        if (res == ESS_RESULT_SUCCESS)
            PRINT(("dev[%u] HAL: %s", i, bufstr));
        res = essIoctl(h, ESS_IOCTL_GET_DEV_DRIVERINFOSTR, bufstr, (uint32_t)sizeof(bufstr));
        if (res == ESS_RESULT_SUCCESS)
            PRINT(("dev[%u] Driver: %s", i, bufstr));
        res = essIoctl(h, ESS_IOCTL_GET_DEV_PRODNAMESTR, bufstr, (uint32_t)sizeof(bufstr));
        if (res == ESS_RESULT_SUCCESS)
            PRINT(("dev[%u] Product Name: %s", i, bufstr));
        if (4 == essESCRead8(h, ESC_REG_TYPE)) {
            if (0x17 == (essESCRead32(h, ESC_REG_IPCORE_VENDORID) & 0xFFFFFF)) {
                uint32_t prodid[3] = { 0, 0, 0 };
                uint8_t* pId = (uint8_t*)prodid;
                int j;

                /* Product ID (as ASCII text) for ESD IP Cores */
                prodid[0] = essESCRead32(h, ESC_REG_IPCORE_PRODID);
                prodid[1] = essESCRead32(h, ESC_REG_IPCORE_PRODID + 4);

                for (j = 0; j < 8; j++, pId++) {
                    if (!isprint((int)*pId)) {
                        *pId = '.';
                    }
                }
                PRINT(("dev[%u] Prod-ID: %s", i, (char*)prodid));
            }
        }
        res = essIoctl(h, ESS_IOCTL_GET_DEV_ORDERNOSTR, bufstr, (uint32_t)sizeof(bufstr));
        if (res == ESS_RESULT_SUCCESS)
            PRINT(("dev[%u] Order No.: %s", i, bufstr));
        res = essIoctl(h, ESS_IOCTL_GET_DEV_SERIALNOSTR, bufstr, (uint32_t)sizeof(bufstr));
        if (res == ESS_RESULT_SUCCESS) {
            PRINT(("dev[%u] Serial No.: %s", i, bufstr));
        } else {
            res = essIoctl(h, ESS_IOCTL_GET_DEV_SERIALNO32, &buf32, (uint32_t)sizeof(buf32));
            if (res == ESS_RESULT_SUCCESS)
                PRINT(("dev[%u] Serial No.: 0x%.8x", i, buf32));
        }
        res = essIoctl(h, ESS_IOCTL_GET_DEV_HWVERSION32, &buf32, (uint32_t)sizeof(buf32));
        if (res == ESS_RESULT_SUCCESS)
            PRINT(("dev[%u] HW Version: 0x%.8x", i, buf32));
        res = essIoctl(h, ESS_IOCTL_GET_DEV_HWFEATURES32, &buf32, (uint32_t)sizeof(buf32));
        if (res == ESS_RESULT_SUCCESS)
            PRINT(("dev[%u] HW Features: 0x%.8x", i, buf32));

        res = essIoctl(h, ESS_IOCTL_GET_DEV_TEMP16, &buf32, (uint32_t)sizeof(buf32));
        if (res == ESS_RESULT_SUCCESS)
        {
        	uint16_t temperature;
        	
        	memcpy(&temperature, &buf32, 2);
            PRINT(("dev[%u] Temperature: %u.%u deg. C", i, 
            		((temperature * 10) >> 4) / 10, (((temperature) * 10) >> 4) % 10));
        }

        essClose(h);

    }
}


#if defined(ESS_PLATFORM_VXWORKS) || defined (ESS_PLATFORM_STM32CUBE)
int essComplexStop(void)
{
	return essStop(hDev);
}

int essComplexStart(int devIdx) {
    ESS_RESULT res;

# else
int ESS_COMPLEX_MAIN(int argc, char *argv[])
{
    ESS_RESULT       res;
    ESS_DEVICE_INDEX devIdx = 0;
    
    if(argc > 1) {
        devIdx = (ESS_DEVICE_INDEX)atoi(argv[1]);
    }

# endif
    
    /* Do platform specific initializations etc.: */
    InitApplication();

#ifdef EEPROM_FILE_NAME
    {
        FILE* fp;

        /* Read emulated EEPROM data from file */
        fp = fopen(EEPROM_FILE_NAME, "rb");
        if (fp != NULL) {
            /* Determine the file size */
            (void)fseek(fp, 0, SEEK_END);
            if (ftell(fp) != sizeof(eepromData)) {
                fclose(fp);
                PRINT(("EEPROM binary must have %d bytes instead of %d",
                        (int)sizeof(eepromData), ftell(fp)));
                return 5;
            }
            rewind(fp);
            if (fread(eepromData, sizeof(eepromData), 1, fp) != 1) {
                fclose(fp);
                PRINT(("Failed to read EEPROM data"));
                return 5;
            }
            fclose(fp);
        }
        else {
            PRINT(("Failed to open EEPROM binary '%s'", EEPROM_FILE_NAME));
            return 5;
        }
    }
#endif

    /* Show info about used soft- and hardware: (useful for support) */
    ShowVersionInfos();

    /* Open ESC and initialize stack:, resulting handle will be copied to hDev */
    res = essOpen(devIdx, &config, &hDev);
    if (res != ESS_RESULT_SUCCESS) {
        PRINT(("Opening device %d failed with %s", devIdx, essFormatResult(res)));
        return 10;
    }

    /* Init object dictionary etc.: */
    Init();

    /* Disable unwanted interrupts: */
    InitALEventMask();

#if (CFG_ESS_EXPLICIT_DEVICE_ID)
    /* Initialize Device Identification Value from ID-Selector here */
    devIdValue = 0x0005;  /* Note: This is the expected value of the CTT 2.x */
#endif

    PRINT(("dev[%u] Starting application ...", devIdx));
    res = essStart(hDev);
    if (res != ESS_RESULT_SUCCESS)
        PRINT(("essStart() returned %s", essFormatResult(res)));

    essClose(hDev);
    return 0;
}

/* Just to make sure the source is used as it should: */
#ifdef ESS_HAL_INCLUDED
# error essHAL.h should not be included by Application!
#endif

#ifdef ESS_PRIVATE_INCLUDED
# error essPrivate.h should not be included by Application!
#endif
