#include "../essHAL.h"


typedef struct {
    int opened;
    /* ... dev specific state infos */
} DEV_INFOS;

static DEV_INFOS devInfos[CFG_ESS_MAX_DEVICES];

#if (CFG_ESS_MAX_DEVICES == 0)
# error ...
#endif


static int global_init_done;
static int devices_opened;
/* Just to handle global_init() and global_finish()  (No thread safety required: parallel access to essOpen() and essClose() is impossible by definition) */

/* Called for first essHALOpen() call. Might be called multiple times if opening is unsuccessful */
static int global_init(void)
{
    if (global_init_done == 0) {
        global_init_done = 1;

        /* ... */
    }

    return 0;
}

/* Called for last essHALClose() call */
static void global_finish(void)
{
    if (global_init_done) {
        /* ... */
        global_init_done = 0;
    }
}



/* Called by essOpen(), no other HAL function for this devIdx is used before */
HAL_EXPORT ESS_RESULT HAL_CALLTYPE essHALOpen(ESS_DEVICE_INDEX devIdx)
{
    DEV_INFOS* dev;

    if (devIdx >= CFG_ESS_MAX_DEVICES)
        return ESS_RESULT_INVALID_INDEX;

    if (devices_opened == 0)
        if (global_init())
            return ESS_RESULT_UNKNOWN;

    dev = &devInfos[devIdx];
    if (dev->opened)
        return ESS_RESULT_ALREADY_ACTIVE;

    if (0) {
        /* ... */
        devices_opened++;

        return ESS_RESULT_SUCCESS;

    } else {
        return ESS_RESULT_NOT_IMPLEMENTED;
    }
}

/* Called by essClose() for a device */
HAL_EXPORT ESS_RESULT HAL_CALLTYPE essHALClose(ESS_DEVICE_INDEX devIdx)
{
    DEV_INFOS* dev;

    if (devIdx >= CFG_ESS_MAX_DEVICES)
        return ESS_RESULT_INVALID_INDEX;

    dev = &devInfos[devIdx];

    if (0) {
        /* ... */
        (void)dev;

        /* essClose() includes call to essHALStop()... */

        devices_opened--;
        if (devices_opened == 0)
            global_finish();

        return ESS_RESULT_SUCCESS;

    } else {
        return ESS_RESULT_NOT_IMPLEMENTED;
    }
}

/* Time in milliseconds (must wrap at 0xffffffff) */
HAL_EXPORT ESS_TIMESTAMP HAL_CALLTYPE essHALGetTime(void)
{
    /* ... */
    return 0;
}

/* Called by essOpen() */
HAL_EXPORT ESS_RESULT HAL_CALLTYPE essHALMapESC(ESS_DEVICE_INDEX devIdx, void** escPointer)
{
    DEV_INFOS* dev;

    if (devIdx >= CFG_ESS_MAX_DEVICES)
        return ESS_RESULT_INVALID_INDEX;

    dev = &devInfos[devIdx];

    /* ... */
    (void)dev;
    (void)escPointer;

    return ESS_RESULT_NOT_IMPLEMENTED;
}

/* Called by essClose() */
HAL_EXPORT ESS_RESULT HAL_CALLTYPE essHALUnmapESC(ESS_DEVICE_INDEX devIdx)
{
    DEV_INFOS* dev;

    if (devIdx >= CFG_ESS_MAX_DEVICES)
        return ESS_RESULT_INVALID_INDEX;

    dev = &devInfos[devIdx];

    /* ... */
    (void)dev;

    return ESS_RESULT_NOT_IMPLEMENTED;
}

/* Called by essStart(). HAL takes control here and stack only runs by cyclic/isr callback then */
HAL_EXPORT ESS_RESULT HAL_CALLTYPE essHALStart(ESS_DEVICE_INDEX devIdx, ESS_HAL_CALLBACK* isr, ESS_HAL_CALLBACK* cyclicCallback, ESS_TIMESTAMP cyclicInterval, void* cbData)
{
    DEV_INFOS* dev;

    if (devIdx >= CFG_ESS_MAX_DEVICES)
        return ESS_RESULT_INVALID_INDEX;

    dev = &devInfos[devIdx];

    /* ... */
    (void)dev;
    (void)isr;
    (void)cyclicCallback;
    (void)cyclicInterval;
    (void)cbData;

    return ESS_RESULT_NOT_IMPLEMENTED;
}

/* Called by essStop() to abort the loop started by essHALStart() (also called if essHALStart() was never called) */
HAL_EXPORT ESS_RESULT HAL_CALLTYPE essHALStop(ESS_DEVICE_INDEX devIdx)
{
    DEV_INFOS* dev;

    if (devIdx >= CFG_ESS_MAX_DEVICES)
        return ESS_RESULT_INVALID_INDEX;

    dev = &devInfos[devIdx];

    /* ... */
    (void)dev;

    return ESS_RESULT_NOT_IMPLEMENTED;
}

/* Not called if stack was built with CFG_ESC_HAVE_POINTER */
HAL_EXPORT void ESS_CALLTYPE essHALReadESCMem(ESS_DEVICE_INDEX devIdx, uint16_t address, void* dst, uint16_t len)
{
    /* ... */
    (void)devIdx;
    (void)address;
    (void)dst;
    (void)len;
}

/* Not called if stack was built with CFG_ESC_HAVE_POINTER */
HAL_EXPORT void ESS_CALLTYPE essHALWriteESCMem(ESS_DEVICE_INDEX devIdx, uint16_t address, const void* src, uint16_t len)
{
    /* ... */
    (void)devIdx;
    (void)address;
    (void)src;
    (void)len;
}

/* ... */
HAL_EXPORT ESS_RESULT HAL_CALLTYPE essHALSetLEDState(ESS_DEVICE_INDEX devIdx, ESC_LED_TYPE ledId, ESS_BOOL ledOn)
{
    DEV_INFOS* dev;

    if (devIdx >= CFG_ESS_MAX_DEVICES)
        return ESS_RESULT_INVALID_INDEX;

    dev = &devInfos[devIdx];

    /* ... */
    (void)dev;
    (void)ledId;
    (void)ledOn;

    return ESS_RESULT_NOT_IMPLEMENTED;
}

/* Only called if stack build with CFG_HAL_NEEDS_EVENTS (Common ESCs such as ET1100/Beckhoff IP do not need this) */
HAL_EXPORT void HAL_CALLTYPE essHALStackEvent(ESS_EVENT_DATA* e)
{
    (void)e;
}

/* Use ESS_IOCTLs starting at 0x80000000 for your own IOCTLs that will never be used by esd itself. */
HAL_EXPORT ESS_RESULT HAL_CALLTYPE essHALIoctl(ESS_DEVICE_INDEX devIdx, ESS_IOCTL fn, void* data, uint32_t sizeOfData)
{
    DEV_INFOS* dev;

    if (devIdx >= CFG_ESS_MAX_DEVICES)
        return ESS_RESULT_INVALID_INDEX;

    dev = &devInfos[devIdx];

    (void)dev;
    (void)data;
    (void)sizeOfData;

    switch (fn) {
        case ESS_IOCTL_GET_DEV_DRIVERINFOSTR:
            /* This might be implemented if it's possible to get some infos about the device driver... */
            /* For example Windows HAL copies something like "esd PCIe EtherCAT Slave Driver, Version 1.0.3, built at Nov 17 2014 05:53:56, DEBUG, ECS-PCIe/1100" */
            return ESS_RESULT_NOT_IMPLEMENTED;

        case ESS_IOCTL_GET_HAL_BUILDINFOSTR:
            /* This might be implemented if HAL is built separately from stack (CFG_HAL_EXTERNAL) */
            return ESS_RESULT_NOT_IMPLEMENTED;

        /* ... */

        default:
            return ESS_RESULT_NOT_IMPLEMENTED;
    }
}
