/*
*       Copyright (c) 2011-2020 by esd electronics gmbh
*
*  This software is copyrighted by and is the sole property of
*  esd gmbh.  All rights, title, ownership, or other interests
*  in the software remain the property of esd gmbh. This
*  software may only be used in accordance with the corresponding
*  license agreement.  Any unauthorized use, duplication, transmission,
*  distribution, or disclosure of this software is expressly forbidden.
*
*  This Copyright notice may not be removed or modified without prior
*  written consent of esd gmbh.
*
*  esd gmbh, reserves the right to modify this software without notice.
*
*  esd electronics gmbh                   Tel. +49-511-37298-0
*  Vahrenwalder Strasse 207               Fax. +49-511-37298-68
*  30165 Hannover                         http://www.esd.eu
*  Germany                                sales@esd.eu
*
*************************************************************************
*
* Description:
* =============================================================================
* This is the file that has to be included to use the esd EtherCAT slave stack.
* (Not used by HAL)
*
*/


#ifndef ESS_INCLUDED
#define ESS_INCLUDED


#include "essSystem.h"
#include "essTypes.h"
#include "ecatDefs.h"



#ifdef __cplusplus
extern "C" {
#endif

                                /*******************************************/
                                /* General functions: */
                                /*******************************************/

                                /* Returns slave stack version number, e.g. "Version 1.0.0, build Apr 21 2011 10:24:56" */
    ESS_EXPORT const char*      ESS_CALLTYPE essGetVersion(void);

                                /* Returns timestamp in milliseconds */
    ESS_EXPORT ESS_TIMESTAMP    ESS_CALLTYPE essGetTime(void);

                                /* Returns a string for the ESS_RESULT code, e.g. "ESS_RESULT_NO_MEM" */
    ESS_EXPORT const char*      ESS_CALLTYPE essFormatResult(ESS_RESULT err);

                                /* "config" is not copied */
    ESS_EXPORT ESS_RESULT       ESS_CALLTYPE essOpen(ESS_DEVICE_INDEX devIdx, const ESS_CONFIGURATION* config, ESS_HANDLE* resHandle);

                                /* Closes the ess handle retrieved by essOpen() */
    ESS_EXPORT ESS_RESULT       ESS_CALLTYPE essClose(ESS_HANDLE h);

                                /* Starts stack's main loop */
    ESS_EXPORT ESS_RESULT       ESS_CALLTYPE essStart(ESS_HANDLE h);

                                /* Stops stack's main loop */
    ESS_EXPORT ESS_RESULT       ESS_CALLTYPE essStop(ESS_HANDLE h);

                                /* Writes to ESC_REG_ALSTATUSCODE and ESC_REG_ALSTATUS. Should be used when an error was detected outside state transition callback */
    ESS_EXPORT ESS_RESULT       ESS_CALLTYPE essIndicateError(ESS_HANDLE h, ESC_STATE newState, REG_VAL_ALSTATUSCODE errCode);

                                /* Sends a CoE Emergency message (Mailbox service). (content at "data" is copied) (Result can be ESS_RESULT_TRY_AGAIN when there's still a queued emergency) */
    ESS_EXPORT ESS_RESULT       ESS_CALLTYPE essCoESendEmergency(ESS_HANDLE h, const ESS_COE_EMERGENCY* data);

                                /* This has to be called when an entry that can be mapped into an input PDO was changed, the stack updates the input SM buffers then. (With ESS_ISYN_FLAGS_NONE after the application call is finished, so this can be called multiple times without triggering multiple copy operations. Use ESS_ISYN_FLAGS_SMx_IMMEDIATELY to change this behavior and update the SM buffer immediately) */
    ESS_EXPORT ESS_RESULT       ESS_CALLTYPE essSyncInputs(ESS_HANDLE h, ESS_ISYN_FLAGS flags);

                                /* addr and len in byte (addr must be even). No swapping is done.  */
    ESS_EXPORT ESS_RESULT       ESS_CALLTYPE essEEPROMRead(ESS_HANDLE h, uint32_t addr, uint32_t len, void* dest);

                                /* addr and len in byte (addr and len must be even). No swapping is done. */
    ESS_EXPORT ESS_RESULT       ESS_CALLTYPE essEEPROMWrite(ESS_HANDLE h, uint32_t addr, uint32_t len, const void* src);

                                /* Sets the state of a LED */
    ESS_EXPORT ESS_RESULT       ESS_CALLTYPE essSetLEDState(ESS_HANDLE h, ESC_LED_TYPE led, ESC_LED_STATE state);

                                /* Reads "len" bytes from ESC memory, no swapping is performed */
    ESS_EXPORT void             ESS_CALLTYPE essESCRead(ESS_HANDLE h, uint16_t address, void* dest, uint16_t len);

                                /* Writes "len" bytes to ESC memory, no swapping is performed */
    ESS_EXPORT void             ESS_CALLTYPE essESCWrite(ESS_HANDLE h, uint16_t address, const void* src, uint16_t len);

                                /* Reads a byte from ESC memory */
    ESS_EXPORT uint8_t          ESS_CALLTYPE essESCRead8(ESS_HANDLE h, uint16_t address);

                                /* Reads a word from ESC memory (swapped on big endian systems) */
    ESS_EXPORT uint16_t         ESS_CALLTYPE essESCRead16(ESS_HANDLE h, uint16_t address);

                                /* Reads a long word from ESC memory (swapped on big endian systems) */
    ESS_EXPORT uint32_t         ESS_CALLTYPE essESCRead32(ESS_HANDLE h, uint16_t address);

                                /* Writes a byte to ESC memory */
    ESS_EXPORT void             ESS_CALLTYPE essESCWrite8(ESS_HANDLE h, uint16_t address, uint8_t value);

                                /* Writes a word to ESC memory (swapped on big endian systems) */
    ESS_EXPORT void             ESS_CALLTYPE essESCWrite16(ESS_HANDLE h, uint16_t address, uint16_t value);

                                /* Writes a long word to ESC memory (swapped on big endian systems) */
    ESS_EXPORT void             ESS_CALLTYPE essESCWrite32(ESS_HANDLE h, uint16_t address, uint32_t value);

                                /*******************************************/
                                /* "essOD" functions (CoE): */
                                /*******************************************/
    ESS_EXPORT ESS_RESULT       ESS_CALLTYPE essODCreate(ESS_HANDLE h, ESS_OD_FLAGS flags);

                                /* Completely deletes the dict, so essODCreate() can/must be used again. (essClose() automatically calls this this) */
    ESS_EXPORT ESS_RESULT       ESS_CALLTYPE essODDelete(ESS_HANDLE h, uint16_t reserved);

                                /* "infos" is optional, needed/used for SDO info service */
    ESS_EXPORT ESS_RESULT       ESS_CALLTYPE essODObjectAdd(ESS_HANDLE h, uint16_t idx, ESS_OD_OBJECT_FLAGS flags, const ESS_OD_OBJECT_INFOS* infos);

                                /* Deletes the given object and all its entries */
    ESS_EXPORT ESS_RESULT       ESS_CALLTYPE essODObjectDelete(ESS_HANDLE h, uint16_t idx);

                                /* "infos" is optional, needed/used for SDO info service */
    ESS_EXPORT ESS_RESULT       ESS_CALLTYPE essODEntryAdd(ESS_HANDLE h, uint16_t idx, uint8_t subIdx, uint32_t bitLen, volatile void* data, COE_ACCESS access, ESS_OD_ENTRY_FLAGS flags, const ESS_OD_ENTRY_INFOS* infos);

                                /* Deletes only the entry */
    ESS_EXPORT ESS_RESULT       ESS_CALLTYPE essODEntryDelete(ESS_HANDLE h, uint16_t idx, uint8_t subIdx);

                                /* When "entriesData" is NULL callback is used */
    ESS_EXPORT ESS_RESULT       ESS_CALLTYPE essODAddArrayObject(ESS_HANDLE h, uint16_t idx, const char* name, uint8_t entriesCount, COE_DATATYPE entriesDataType, uint32_t entryBitLen, COE_ACCESS entriesAcc, volatile void* entriesData, uint16_t entriesDataStride);

                                /* Define a virtual max entry size for RO arrays */
    ESS_EXPORT ESS_RESULT       ESS_CALLTYPE essODSetArrayMax(ESS_HANDLE h, uint16_t idx, uint8_t entriesMax);

                                /* Adds some generic objects (All Readonly). All numbers must be stored as little endian. "deviceType": mandatory, Object 0x1000 (Device Type). "vendorId", "productCode", "revNo", "serialNo": mandatory, Object 0x1018 (Identity Object). "devName": optional, Object 0x1008 (Manufacturer Device name). "verHardware: optional, Object 0x1009 (Manufacturer Hardware Version). "verSoftware: optional, Object 0x100a (Manufacturer Software Version) */
    ESS_EXPORT ESS_RESULT       ESS_CALLTYPE essODAddGenericObjects(ESS_HANDLE h, const uint32_t* deviceType, const uint32_t* vendorId, const uint32_t* productCode, const uint32_t* revNo, const uint32_t* serialNo, const char* devName, const char* verHardware, const char* verSoftware);

                                /* Updates a PDO configuration (objects 0x1600/0x1a00, etc.) "writable": SDO write access in PreOp */
    ESS_EXPORT ESS_RESULT       ESS_CALLTYPE essODUpdatePDOConfiguration(ESS_HANDLE h, uint16_t pdo, const char* name, const ESS_PDO_ENTRY* entries, uint8_t entryCount, ESS_PDO_CFG_FLAGS flags);

                                /* When "dest" is null, number of entries in the given PDO is returned in "numEntries". Else PDO entries are copied to "dest": "numEntries" in = max items to copy to "dest", out = total number of items written to "dest" (reserved must be set 0) */
    ESS_EXPORT ESS_RESULT       ESS_CALLTYPE essODGetPDOConfiguration(ESS_HANDLE h, uint16_t pdo, uint8_t* numEntries, ESS_PDO_ENTRY* dest, uint8_t reserved);

                                /* Updates SM PDO assignment (objects 0x1c12/0x1c13, etc.) "writable": SDO write access in PreOp */
    ESS_EXPORT ESS_RESULT       ESS_CALLTYPE essODUpdatePDOAssignment(ESS_HANDLE h, ESS_SM smIdx, const uint16_t* pdo, uint8_t pdoCount, ESS_BOOL writable);

                                /* When "dest" is null, number of PDOs assigned to the given SM is returned in "numPDOs". Else PDO indices are copied to "dest": "numPDOs" in = max items to copy to "dest", out = total number of items written to "dest" */
    ESS_EXPORT ESS_RESULT       ESS_CALLTYPE essODGetPDOAssignment(ESS_HANDLE h, ESS_SM smIdx, uint8_t* numPDOs, uint16_t* dest);

                                /* Creates the Parameter object for a PDO (Obj 0x1801 for PDO 0x1a01, etc.) */
    ESS_EXPORT ESS_RESULT       ESS_CALLTYPE essODPDOParamCreate(ESS_HANDLE h, uint16_t pdo, const char* name, ESS_OD_PDOPARAM_FLAGS flags);

                                /* Updates value of subindex 6 of the according PDO Parameter object. "data" as needed in the dict, e.g. "\x00\x1a\x01\x1a" when the PDOs 0x1a00 and 0x1a01 are excluded. */
    ESS_EXPORT ESS_RESULT       ESS_CALLTYPE essODPDOParamUpdateExclude(ESS_HANDLE h, uint16_t pdo, const void* data, uint16_t dataLen);

                                /* Gets the pointer to entry 7 of the according PDO Parameter object. (Entry is used to indicate inputs invalid or outputs can't be written to hardware.) */
    ESS_EXPORT ESS_RESULT       ESS_CALLTYPE essODPDOParamGetState(ESS_HANDLE h, uint16_t pdo, volatile ESS_BOOL** state);

                                /* Gets the pointer to entry 8 of the according PDO Parameter object. (Entry used for output PDOs, master sets this to true when output values are invalid.) */
    ESS_EXPORT ESS_RESULT       ESS_CALLTYPE essODPDOParamGetControl(ESS_HANDLE h, uint16_t pdo, volatile ESS_BOOL** control);

                                /* Gets the pointer to entry 9 of the according PDO Parameter object (Entry is used/toggled to indicate changed inputs/outputs.) */
    ESS_EXPORT ESS_RESULT       ESS_CALLTYPE essODPDOParamGetToggle(ESS_HANDLE h, uint16_t pdo, volatile ESS_BOOL** toggle);


                                /*******************************************/
                                /* EoE functions: */
                                /*******************************************/

                                /* Sends an Ethernet frame to EtherCAT. "reserved" must be NULL */
    ESS_EXPORT ESS_RESULT       ESS_CALLTYPE essEoESendFrame(ESS_HANDLE h, const void* data, uint16_t dataLen, void* reserved);


                                /*******************************************/
                                /* Other functions: */
                                /*******************************************/

                                /* Returns tag given with ESS_CONFIGURATION */
    ESS_EXPORT void*            ESS_CALLTYPE essGetTag(ESS_HANDLE h);

                                /* To implement other, special, functions. Should be needed only for specific hardware/targets - i.e. most IOCTLs are not implemented on most targets and essIoctl() will just return ESS_RESULT_NOT_IMPLEMENTED */
    ESS_EXPORT ESS_RESULT       ESS_CALLTYPE essIoctl(ESS_HANDLE h, ESS_IOCTL fn, void* data, uint32_t sizeOfData);

#ifdef __cplusplus
} /* extern "C" */
#endif


/* Useful macros: */
/******************/
#define LO_DWORD(v64)                                   ((uint32_t)((v64) & 0xffffffff))
#define HI_DWORD(v64)                                   ((uint32_t)(((v64) >> 32) & 0xffffffff))
#define LO_WORD(v32)                                    ((uint16_t)((v32) & 0xffff))
#define HI_WORD(v32)                                    ((uint16_t)(((v32) >> 16) & 0xffff))
#define LO_BYTE(v16)                                    ((uint8_t)((v16) & 0xff))
#define HI_BYTE(v16)                                    ((uint8_t)(((v16) >> 8) & 0xff))

#if !defined(SWAP16) || !defined(SWAP32)
# if defined(__QNXNTO__)
#  define SWAP16(v16)                                   ENDIAN_RET16(v16)
#  define SWAP32(v32)                                   ENDIAN_RET32(v32)
# elif defined(__GNUC__) && defined(__i486__) && !defined(__STRICT_ANSI__)
#  define SWAP16(v16)                                   ({ register uint16_t __x; asm ("xchg %b1,%h1" : "=Q" (__x) : "0" ((uint16_t)v16)); __x; })
#  define SWAP32(v32)                                   ({ register uint32_t __x; asm ("bswap %1" : "=r" (__x) : "0" ((uint32_t)v32)); __x; })
# elif defined(_MSC_VER)
#  define SWAP16(v16)                                   _byteswap_short(v16)
#  define SWAP32(v32)                                   _byteswap_ulong(v32)
# else
#  define SWAP16(v16)                                   ((uint16_t)(((uint16_t)(v16) >> 8) | ((uint16_t)(v16) << 8)))
#  define SWAP32(v32)                                   ((uint32_t)(((uint32_t)(v32) >> 24) | ((uint32_t)(v32) << 24) | (((uint32_t)(v32) >> 8) & 0x0000ff00) | (((uint32_t)(v32) << 8) & 0x00ff0000)))
# endif
#endif
#if !defined(SWAP64)
# if defined(__QNXNTO__)
#  define SWAP64(v64)                                   ENDIAN_RET64(v64)
# elif defined(_MSC_VER)
#  define SWAP64(v64)                                   _byteswap_uint64(v64)
# else
#  define SWAP64(v64)                                   ((SWAP32((v64) >> 32)) | ((uint64_t)(SWAP32(v64)) << 32))
# endif
#endif
#if !defined(SWAP_WORD)
# define SWAP_WORD(a, b)                                do { uint16_t _temp = (b); (b) = (a); (a) = _temp; } while (0)
#endif
#if !defined(MIN) || !defined(MAX)
# define MIN(a, b)                                      (((a) < (b)) ? (a) : (b))
# define MAX(a, b)                                      (((a) > (b)) ? (a) : (b))
#endif
#define MAKE_WORD(loByte, hiByte)                       (uint16_t)((uint16_t)(loByte) | ((uint16_t)(hiByte) << 8))
#define MAKE_DWORD(loWLoB, loWHiB, hiWLoB, hiWHiB)      (((uint32_t)(loWLoB)) | ((uint32_t)(loWHiB) << 8) | ((uint32_t)(hiWLoB) << 16) | ((uint32_t)(hiWHiB) << 24))
#define MAKE_ESS_BOOL(c)                                ((c) ? (ESS_BOOL)1 : (ESS_BOOL)0)
#define MAKE_BIT_MASK(firstBit, bitLen)                 (((1U << (bitLen)) - 1) << (firstBit))
#define BIT_FROM_BYTEARR(pBytes, bitPos)                ((((pBytes)[(bitPos) >> 3] & (1 << ((bitPos) % 8))) != 0))
#define BIT_TO_BYTEARR(pBytes, bitPos, val)             do { if (val) (pBytes)[(bitPos) >> 3] |= 1 << ((bitPos) % 8); else (pBytes)[(bitPos) >> 3] &= ~(1 << ((bitPos) % 8)); } while (0)
#define BYTE_FROM_BYTEARR(pByte, bitPos)                (((pByte)[(bitPos) >> 3] >> ((bitPos) % 8)) | ((pByte)[((bitPos) >> 3) + 1] << (8 - ((bitPos) % 8))))

#if CFG_BIG_ENDIAN
# define SWAP16_LE2HOST(v16)                            SWAP16(v16)
# define SWAP16_HOST2LE(v16)                            SWAP16(v16)
# define SWAP32_LE2HOST(v32)                            SWAP32(v32)
# define SWAP32_HOST2LE(v32)                            SWAP32(v32)
# define SWAP64_LE2HOST(v64)                            SWAP64(v64)
# define SWAP64_HOST2LE(v64)                            SWAP64(v64)
#else
# define SWAP16_LE2HOST(v16)                            (v16)
# define SWAP16_HOST2LE(v16)                            (v16)
# define SWAP32_LE2HOST(v32)                            (v32)
# define SWAP32_HOST2LE(v32)                            (v32)
# define SWAP64_LE2HOST(v64)                            (v64)
# define SWAP64_HOST2LE(v64)                            (v64)
#endif
                                                        /* 64 bit access only as macro because uint64_t type is not crucial / may not exist. */
#define essESCRead64To(h, addr, ptrToAligned64)         do { uint64_t _temp; essESCRead((h), (addr), &_temp, 8); *(ptrToAligned64) = SWAP64_LE2HOST(_temp); } while (0)
#define essESCWrite64(h, addr, v64)                     do { uint64_t _temp = (v64); _temp = SWAP64_HOST2LE(_temp); essESCWrite((h), (addr), &_temp, 8); } while (0)

#if (!CFG_ESS_DISABLE_DEPRECATED)
# define essODSyncInputMappedEntries(h)                 essSyncInputs((h), ESS_ISYN_FLAGS_NONE)
#endif

#endif /* ESS_INCLUDED */
